from __future__ import annotations

import sys
from typing import TYPE_CHECKING

import numpy as np
import pyomo.environ as pyo
from pyomo import opt
from stable_baselines3.common.base_class import BaseAlgorithm
from stable_baselines3.common.vec_env import VecEnv, VecNormalize

from eta_utility import get_logger
from eta_utility.util import deprecated

if TYPE_CHECKING:
    import io
    import pathlib
    from collections.abc import Sequence
    from typing import Any

    import torch as th
    from stable_baselines3.common.policies import BasePolicy
    from stable_baselines3.common.type_aliases import GymEnv, MaybeCallback

log = get_logger("eta_x.agents")


class MathSolver(BaseAlgorithm):
    """Simple, Pyomo based MPC agent.

    The agent requires an environment that specifies the 'model' attribute, returning a
    :py:class:`pyomo.ConcreteModel` and a sorted list as the order for the action space. This list is used to
    avoid ambiguity when returning a list of actions. Since the model specifies its own action and observation
    space, this agent does not use the *action_space* and *observation_space* specified by the environment.

    :param policy: Agent policy. Parameter is not used in this agent.
    :param env: Environment to be optimized.
    :param verbose: Logging verbosity.
    :param solver_name: Name of the solver, could be cplex or glpk.
    :param action_index: Index of the solution value to be used as action (if this is 0, the first value in a list
        of solution values will be used).
    :param kwargs: Additional arguments as specified in stable_baselins3.commom.base_class or as provided by solver.
    """

    def __init__(
        self,
        policy: type[BasePolicy],
        env: VecEnv,
        verbose: int = 1,
        *,
        solver_name: str = "cplex",
        action_index: int = 0,
        _init_setup_model: bool = True,
        **kwargs: Any,
    ) -> None:
        # Prepare kwargs to be sent to the super class and to the solver.
        super_args = {}
        solver_args = {}

        # Set default values for superclass arguments
        kwargs.setdefault("learning_rate", 0)

        for key in kwargs:
            # Find arguments which are meant for the BaseAlgorithm class and extract them into super_args
            if key in {
                "policy_base",
                "learning_rate",
                "policy_kwargs",
                "device",
                "support_multi_env",
                "create_eval_env",
                "monitor_wrapper",
                "seed",
                "use_sde",
                "sde_sample_freq",
                "supported_action_spaces",
            }:
                super_args[key] = kwargs.get(key, None)
            elif key in {"tensorboard_log"}:
                log.warning(
                    "The MPC Basic agent does not support logging to tensorboard. "
                    "Ignoring parameter tensorboard_log."
                )
            else:
                solver_args[key] = kwargs[key]

        super().__init__(policy=policy, env=env, verbose=verbose, **super_args)
        log.setLevel(int(verbose * 10))  # Set logging verbosity

        # Check configuration for MILP compatibility
        if self.env is None:
            raise ValueError("The MPC agent needs a specific environment to work correctly. Cannot use env = None.")

        if self.n_envs is not None and self.n_envs > 1:
            raise ValueError(
                "The MPC agent can only use one environment. It cannot work on multiple vectorized environments."
            )
        if isinstance(self.env, VecNormalize):
            raise TypeError("The MPC agent does not allow the use of normalized environments.")

        # Solver parameters
        self.solver_name: str = solver_name
        self.solver_options: dict = {}
        self.solver_options.update(solver_args)

        self.model: pyo.ConcreteModel  #: Pyomo optimization model as specified by the environment.
        self.actions_order: Sequence[str]  #: Specification of the order in which action values should be returned.

        self.policy_class: type[BasePolicy]
        if _init_setup_model:
            self._setup_model()

        #: Index of the solution value to be used as action (if this is 0, the first value in a list
        #: of solution values will be used).
        self.action_index = action_index

    def _setup_model(self) -> None:
        """Load the MILP model from the environment."""
        assert self.env is not None, "The MPC agent needs a specific environment to work. Cannot use env = None."
        self.model, self.actions_order = self.env.get_attr("model", 0)[0]
        if self.policy_class is not None:
            self.policy: type[BasePolicy] = self.policy_class(  # type: ignore
                self.observation_space,
                self.action_space,
            )

    def solve(self) -> pyo.ConcreteModel:
        """Solve the current pyomo model instance with given parameters. This could also be used separately to solve
        normal MILP problems. Since the entire problem instance is returned, result handling can be outsourced.

        :return: Solved pyomo model instance.
        """
        assert self.env is not None, "The MPC agent needs a specific environment to work. Cannot use env = None."

        solver = pyo.SolverFactory(self.solver_name)
        solver.options.update(self.solver_options)  # Adjust solver settings

        _tee: bool = bool(log.level / 10 <= 1)
        result = solver.solve(self.model, symbolic_solver_labels=True, tee=_tee)
        if _tee:
            print("\n")  # noqa: T201 (print is ok here, because cplex prints directly to console).
        log.debug(
            "Problem information: \n"
            "\t+----------------------------------+\n"
            + "\n".join(
                f"\t {item}: {value.value} "
                for item, value in result["Problem"][0].items()
                if not isinstance(value.value, opt.UndefinedData)
            )
            + "\n\t+----------------------------------+"
        )

        # Log status after the optimization
        log.info(
            "Solver information: \n"
            "\t+----------------------------------+\n"
            + "\n".join(
                f"\t {item}: {value.value} "
                for item, value in result["Solver"][0].items()
                if item != "Statistics" and not isinstance(value.value, opt.UndefinedData)
            )
            + "\n\t+----------------------------------+"
        )

        # Log status after the optimization
        if len(result["Solution"]) >= 1:
            log.debug(
                "Solution information: \n"
                "\t+----------------------------------+\n"
                + "\n".join(
                    f"\t {item}: {value.value} "
                    for item, value in result["Solution"][0].items()
                    if not isinstance(value.value, opt.UndefinedData)
                )
                + "\n\t+----------------------------------+"
            )

        # Interrupt execution if no optimal solution could be found
        if (
            result.solver.termination_condition != opt.TerminationCondition.optimal
            or result.solver.status != opt.SolverStatus.ok
        ):
            log.error("Problem can not be solved - aborting.")
            self.env.env_method("solve_failed", self.model, result)
            sys.exit(1)

        return self.model

    def predict(
        self,
        observation: np.ndarray | dict[str, np.ndarray],
        state: tuple[np.ndarray, ...] | None = None,
        episode_start: np.ndarray | None = None,
        deterministic: bool = False,
    ) -> tuple[np.ndarray, tuple[np.ndarray, ...] | None]:
        """
        Solve the current pyomo model instance with given parameters and observations and return the optimal actions.

        :param observation: the input observation (not used here).
        :param state: The last states (not used here).
        :param episode_start: The last masks (not used here).
        :param deterministic: Whether to return deterministic actions. This agent always returns
                                   deterministic actions.
        :return: Tuple of the model's action and the next state (not used here).
        """
        assert self.env is not None, "The MPC agent needs a specific environment to work. Cannot use env = None."
        self.model, _ = self.env.get_attr("model", 0)[0]
        self.solve()
        self.env.set_attr("model", self.model, 0)

        # Aggregate the agent actions from pyomo component objects
        solution = {}
        for com in self.model.component_objects(pyo.Var):
            if isinstance(com, pyo.ScalarVar):
                continue
            try:
                solution[com.name] = pyo.value(com[com.index_set().at(self.action_index + 1)])
            except ValueError:
                pass

        # Make sure that actions are returned to the correct order and as a numpy array.
        actions: np.ndarray = np.ndarray((1, len(self.actions_order)))
        for i, action in enumerate(self.actions_order):
            log.debug(f"Action '{action}' value: {solution[action]}")
            actions[0][i] = solution[action]

        return actions, state

    def action_probability(
        self,
        observation: np.ndarray,
        state: np.ndarray | None = None,
        mask: np.ndarray | None = None,
        actions: np.ndarray | None = None,
        logp: bool = False,
    ) -> None:
        """The MPC approach cannot predict probabilities of single actions."""
        raise NotImplementedError("The MPC agent cannot predict probabilities of single actions.")

    def learn(
        self,
        total_timesteps: int,
        callback: MaybeCallback = None,
        log_interval: int = 100,
        tb_log_name: str = "run",
        reset_num_timesteps: bool = True,
        progress_bar: bool = False,
    ) -> MathSolver:
        """The MPC approach cannot learn a new model. Specify the model attribute as a pyomo Concrete model instead,
        to use the prediction function of this agent.

        :param total_timesteps: The total number of samples (env steps) to train on
        :param callback: callback(s) called at every step with state of the algorithm.
        :param log_interval: The number of timesteps before logging.
        :param tb_log_name: the name of the run for TensorBoard logging
        :param reset_num_timesteps: whether or not to reset the current timestep number (used in logging)
        :param progress_bar: Display a progress bar using tqdm and rich.
        :return: The trained model.
        """

        return self

    @classmethod
    def load(
        cls,
        path: str | pathlib.Path | io.BufferedIOBase,
        env: GymEnv | None = None,
        device: th.device | str = "auto",
        custom_objects: dict[str, Any] | None = None,
        print_system_info: bool = False,
        force_reset: bool = True,
        **kwargs: Any,
    ) -> MathSolver:
        """
        Load the model from a zip-file.
        Warning: ``load`` re-creates the model from scratch, it does not update it in-place!

        :param path: path to the file (or a file-like) where to load the agent from
        :param env: the new environment to run the loaded model on
            (can be None if you only need prediction from a trained model) has priority over any saved environment
        :param device: Device on which the code should run.
        :param custom_objects: Dictionary of objects to replace upon loading. If a variable is present in
            this dictionary as a key, it will not be deserialized and the corresponding item will be used instead.
        :param print_system_info: Whether to print system info from the saved model and the current system info
            (useful to debug loading issues)
        :param force_reset: Force call to ``reset()`` before training to avoid unexpected behavior.
        :param kwargs: extra arguments to change the model when loading
        :return: new model instance with loaded parameters
        """
        if env is None:
            raise ValueError("Parameter env must be specified.")
        model: MathSolver = super().load(path, env, device, custom_objects, print_system_info, force_reset, **kwargs)

        return model


@deprecated("Use `MathSolver` instead of `MPCBasic`.")
class MPCBasic(MathSolver):
    """Deprecated MPCBasic class. Use `MathSolver` instead."""



from __future__ import annotations

import os
import pathlib
from contextlib import contextmanager
from datetime import datetime
from typing import TYPE_CHECKING

import numpy as np
from stable_baselines3.common.callbacks import CheckpointCallback
from stable_baselines3.common.vec_env import VecNormalize

from eta_utility import get_logger
from eta_utility.eta_x import ConfigOpt, ConfigOptRun
from eta_utility.eta_x.common import (
    CallbackEnvironment,
    initialize_model,
    is_env_closed,
    load_model,
    log_net_arch,
    log_run_info,
    log_to_file,
    merge_callbacks,
    vectorize_environment,
)

if TYPE_CHECKING:
    from collections.abc import Generator, Mapping
    from typing import Any

    from stable_baselines3.common.base_class import BaseAlgorithm
    from stable_baselines3.common.type_aliases import MaybeCallback
    from stable_baselines3.common.vec_env import VecEnv
    from stable_baselines3.common.vec_env.base_vec_env import VecEnvObs

log = get_logger("eta_x")


class ETAx:
    """Initialize an optimization model and provide interfaces for optimization, learning and execution (play).

    :param root_path: Root path of the eta_x application (the configuration will be interpreted relative to this).
    :param config_name: Name of configuration .ini file in configuration directory (should be JSON format).
    :param config_overwrite: Dictionary to overwrite selected configurations.
    :param relpath_config: Relative path to configuration file, starting from root path.
    """

    def __init__(
        self,
        root_path: str | os.PathLike,
        config_name: str,
        config_overwrite: Mapping[str, Any] | None = None,
        relpath_config: str | os.PathLike = "config/",
    ) -> None:
        # Load configuration for the optimization
        _root_path = root_path if isinstance(root_path, pathlib.Path) else pathlib.Path(root_path)
        _relpath_config = relpath_config if isinstance(relpath_config, pathlib.Path) else pathlib.Path(relpath_config)
        #: Path to the configuration file.
        self.path_config = _root_path / _relpath_config / f"{config_name}"
        #: ConfigOpt object for the optimization run.
        self.config: ConfigOpt = ConfigOpt.from_config_file(self.path_config, root_path, config_overwrite)
        log.setLevel(int(self.config.settings.verbose * 10))

        #: Configuration for an optimization run.
        self.config_run: ConfigOptRun | None = None

        #: The vectorized environments.
        self.environments: VecEnv | VecNormalize | None = None
        #: Vectorized interaction environments.
        self.interaction_env: VecEnv | None = None
        #: The model or algorithm.
        self.model: BaseAlgorithm | None = None

    @contextmanager
    def prepare_environments_models(
        self,
        series_name: str | None,
        run_name: str | None,
        run_description: str = "",
        reset: bool = False,
        training: bool = False,
    ) -> Generator:
        if is_env_closed(self.environments) or self.model is None:
            _series_name = series_name if series_name is not None else ""
            _run_name = run_name if run_name is not None else ""
            self.prepare_run(_series_name, _run_name, run_description)

        with self.prepare_environments(training):
            assert (
                self.environments is not None
            ), "Initialized environments could not be found. Call prepare_environments first."

            self.prepare_model(reset)
            yield

    def prepare_run(self, series_name: str, run_name: str, run_description: str = "") -> None:
        """Prepare the learn and play methods by reading configuration, creating results folders and the model.

        :param series_name: Name for a series of runs.
        :param run_name: Name for a specific run.
        :param run_description: Description for a specific run.
        """
        self.config_run = ConfigOptRun(
            series=series_name,
            name=run_name,
            description=run_description,
            path_root=self.config.path_root,
            path_results=self.config.path_results,
            path_scenarios=self.config.path_scenarios,
        )
        self.config_run.create_results_folders()

        # Add file handler to parent logger to log the terminal output
        log_to_file(config=self.config, config_run=self.config_run)

        log.info("Run prepared successfully.")

    def prepare_model(self, reset: bool = False) -> None:
        """Check for existing model and load it or back it up and create a new model.

        :param reset: Flag to determine whether an existing model should be reset.
        """
        self._prepare_model(reset=reset)

    def _prepare_model(self, reset: bool = False) -> None:
        """Check for existing model and load it or back it up and create a new model.

        :param reset: Flag to determine whether an existing model should be reset.
        """
        assert self.config_run is not None, (
            "Set the config_run attribute before trying to initialize the model "
            "(for example by calling prepare_run)."
        )
        assert (
            self.environments is not None
        ), "Initialize the environments before trying to initialize the model(for example by calling prepare_run)."

        path_model = self.config_run.path_run_model
        if path_model.is_file() and reset:
            log.info(f"Existing model detected: {path_model}")

            bak_name = path_model / f"_{datetime.fromtimestamp(path_model.stat().st_mtime).strftime('%Y%m%d_%H%M')}.bak"
            path_model.rename(bak_name)
            log.info(f"Reset is active. Existing model will be backed up. Backup file name: {bak_name}")
        elif path_model.is_file():
            log.info(f"Existing model detected: {path_model}. Loading model.")

            self.model = load_model(
                self.config.setup.agent_class,
                self.environments,
                self.config.settings.agent,
                self.config_run.path_run_model,
                tensorboard_log=self.config.setup.tensorboard_log,
                log_path=self.config_run.path_series_results,
            )
            return

        # Initialize the model if it wasn't loaded from a file
        self.model = initialize_model(
            self.config.setup.agent_class,
            self.config.setup.policy_class,
            self.environments,
            self.config.settings.agent,
            self.config.settings.seed,
            tensorboard_log=self.config.setup.tensorboard_log,
            log_path=self.config_run.path_series_results,
        )

    @contextmanager
    def prepare_environments(self, training: bool = True) -> Generator:
        """Context manager which prepares the environments and closes them after it exits.

        :param training: Should preparation be done for training (alternative: playing)?
        """
        # If the agents specifies the population parameter, the number of environments usually has to be
        # equal to that value as well. See NSGA-II agent.
        if (
            "population" in self.config.settings.agent
            and self.config.settings.n_environments != self.config.settings.agent["population"]
        ):
            if self.config.settings.n_environments != 1:
                log.warning(
                    f"Agent specifies 'population' parameter but the number of environments "
                    f"({self.config.settings.n_environments}) is not equal to the population. "
                    f"Setting 'n_environments' to {self.config.settings.agent['population']}"
                )
            self.config.settings.n_environments = self.config.settings.agent["population"]

        try:
            self._prepare_environments(training)
            yield
        finally:
            # close all environments when done (kill processes)
            if self.environments is not None:
                log.debug("Closing environments.")
                self.environments.close()
            else:
                log.error("Environment initialization failed.")

            if self.config.settings.interact_with_env:
                if self.interaction_env is not None:
                    log.debug("Closing interaction environment.")
                    self.interaction_env.close()
                else:
                    log.error("Interaction environment initialization failed.")

    def _prepare_environments(self, training: bool = True) -> None:
        """Vectorize and prepare the environments and potentially the interaction environments.

        :param training: Should preparation be done for training (alternative: playing)?
        """
        # If the agents specifies the population parameter, the number of environments usually has to be
        # equal to that value as well. See NSGA-II agent.
        if (
            "population" in self.config.settings.agent
            and self.config.settings.n_environments != self.config.settings.agent["population"]
        ):
            if self.config.settings.n_environments != 1:
                log.warning(
                    f"Agent specifies 'population' parameter but the number of environments "
                    f"({self.config.settings.n_environments}) is not equal to the population. "
                    f"Setting 'n_environments' to {self.config.settings.agent['population']}"
                )
            self.config.settings.n_environments = self.config.settings.agent["population"]

        assert self.config_run is not None, (
            "Set the config_run attribute before trying to initialize the environments "
            "(for example by calling prepare_run)."
        )

        env_class = self.config.setup.environment_class
        self.config_run.set_env_info(env_class)

        callback = CallbackEnvironment(self.config.settings.plot_interval)
        # Vectorize the environments
        self.environments = vectorize_environment(
            env_class,
            self.config_run,
            self.config.settings.environment,
            callback,
            self.config.settings.verbose,
            self.config.setup.vectorizer_class,
            self.config.settings.n_environments,
            training=training,
            monitor_wrapper=self.config.setup.monitor_wrapper,
            norm_wrapper_obs=self.config.setup.norm_wrapper_obs,
            norm_wrapper_reward=self.config.setup.norm_wrapper_reward,
        )

        if self.config.settings.interact_with_env:
            # Perform some checks to ensure the interaction environment is configured correctly.
            if self.config.setup.interaction_env_class is None:
                raise ValueError(
                    "If 'interact_with_env' is specified, an interaction env class must be specified as well."
                )
            if self.config.settings.interaction_env is None:
                raise ValueError(
                    "If 'interact_with_env' is specified, the interaction_env settings must be specified as well."
                )
            interaction_env_class = self.config.setup.interaction_env_class
            self.config_run.set_interaction_env_info(interaction_env_class)

            # Vectorize the environment
            self.interaction_env = vectorize_environment(
                interaction_env_class,
                self.config_run,
                self.config.settings.interaction_env,
                callback,
                self.config.settings.verbose,
                training=training,
            )

    def learn(
        self,
        series_name: str | None = None,
        run_name: str | None = None,
        run_description: str = "",
        reset: bool = False,
        callbacks: MaybeCallback = None,
    ) -> None:
        """Start the learning job for an agent with the specified environment.

        :param series_name: Name for a series of runs.
        :param run_name: Name for a specific run.
        :param run_description: Description for a specific run.
        :param reset: Indication whether possibly existing models should be reset. Learning will be continued if
                           model exists and reset is false.
        :param callbacks: Provide additional callbacks to send to the model.learn() call.
        """
        with self.prepare_environments_models(series_name, run_name, run_description, reset, training=True):
            assert self.config_run is not None, "Run configuration could not be found. Call prepare_run first."
            assert (
                self.environments is not None
            ), "Initialized environments could not be found. Call prepare_environments first."
            assert self.model is not None, "Initialized model could not be found. Call prepare_model first."

            # Log some information about the model and configuration
            log_net_arch(self.model, self.config_run)
            log_run_info(self.config, self.config_run)

            # Genetic algorithm has a slightly different concept for saving since it does not stop between time steps
            if "n_generations" in self.config.settings.agent:
                save_freq = self.config.settings.save_model_every_x_episodes
                total_timesteps = self.config.settings.agent["n_generations"]
            else:
                # Check if all required config values are present
                if self.config.settings.episode_duration is None:
                    raise ValueError("Missing configuration values for learning: 'episode_duration'.")
                if self.config.settings.sampling_time is None:
                    raise ValueError("Missing configuration values for learning: 'sampling_time'.")
                if self.config.settings.n_episodes_learn is None:
                    raise ValueError("Missing configuration values for learning: 'n_episodes_learn'.")

                # define callback for periodically saving models
                save_freq = int(
                    self.config.settings.episode_duration
                    / self.config.settings.sampling_time
                    * self.config.settings.save_model_every_x_episodes
                )
                total_timesteps = int(
                    self.config.settings.n_episodes_learn
                    * self.config.settings.episode_duration
                    / self.config.settings.sampling_time
                )

            # Set the seed for the environments before starting to learn
            self.environments.seed(self.config.settings.seed)

            callback_learn = merge_callbacks(
                CheckpointCallback(
                    save_freq=save_freq,
                    save_path=str(self.config_run.path_series_results / "models"),
                    name_prefix=self.config_run.name,
                ),
                callbacks,
            )

            # The experiments are reset before the learning phase begins, start learning
            log.info("Start learning process of agent in environment.")
            try:
                self.model.learn(
                    total_timesteps=total_timesteps,
                    callback=callback_learn,
                    tb_log_name=self.config_run.name,
                )
            except OSError:
                filename = str(self.config_run.path_series_results / f"{self.config_run.name}_model_before_error.pkl")
                log.info(f"Saving model to file: {filename}.")
                self.model.save(filename)
                raise

            try:
                log.debug("Resetting environment one more time to call environment callback one last time.")
                self.environments.reset()
            except ValueError as e:
                raise ValueError("An error occured when the environment is resetting.") from e

            # Save model
            log.debug(f"Saving model to file: {self.config_run.path_run_model}.")
            self.model.save(self.config_run.path_run_model)
            if isinstance(self.environments, VecNormalize):
                log.debug(f"Saving environment normalization data to file: {self.config_run.path_vec_normalize}.")
                self.environments.save(str(self.config_run.path_vec_normalize))

        log.info(f"Learning finished: {series_name} / {run_name}")

    def play(self, series_name: str | None = None, run_name: str | None = None, run_description: str = "") -> None:
        """Play with previously learned agent model in environment.

        :param series_name: Name for a series of runs.
        :param run_name: Name for a specific run.
        :param run_description: Description for a specific run.
        """
        with self.prepare_environments_models(series_name, run_name, run_description, reset=False, training=False):
            assert self.config_run is not None, "Run configuration could not be found. Call prepare_run first."
            assert (
                self.environments is not None
            ), "Initialized environments could not be found. Call prepare_environments first."
            assert self.model is not None, "Initialized model could not be found. Call prepare_model first."

            if self.config.settings.n_episodes_play is None:
                raise ValueError("Missing configuration value for playing: 'n_episodes_play' in section 'settings'")

            # Log some information about the model and configuration
            log_net_arch(self.model, self.config_run)
            log_run_info(self.config, self.config_run)

            n_episodes_stop = self.config.settings.n_episodes_play

            # Reset the environments before starting to play
            try:
                log.debug("Resetting environments before starting to play.")
                observations = self._reset_envs()
            except ValueError as e:
                raise ValueError(
                    "It is likely that returned observations do not conform to the specified state config."
                ) from e
            n_episodes = 0

            log.debug("Start playing process of agent in environment.")
            if self.config.settings.interact_with_env:
                log.info("Starting agent with environment/optimization interaction.")
            else:
                log.info("Starting without an additional interaction environment.")

            _round_actions = self.config.settings.round_actions
            _scale_actions = self.config.settings.scale_actions if self.config.settings.scale_actions is not None else 1

            while n_episodes < n_episodes_stop:
                try:
                    observations, dones = self._play_step(_round_actions, _scale_actions, observations)
                except BaseException as e:
                    log.error(
                        "Exception occurred during an environment step. Aborting and trying to reset environments."
                    )
                    try:
                        observations = self._reset_envs()
                    except BaseException as followup_exception:
                        raise e from followup_exception
                    log.debug("Environment reset successful - re-raising exception")
                    raise e

                n_episodes += sum(dones)

    def _play_step(
        self, _round_actions: int | None, _scale_actions: float, observations: VecEnvObs
    ) -> tuple[VecEnvObs, np.ndarray]:
        assert self.environments is not None, "Initialized environments could not be found. Call prepare_run first."

        # set policy prediction to deterministic for playing; type: ignore
        # Type ignored because typing in eta_x is bad
        action, _ = self.model.predict(observation=observations, deterministic=True)  # type: ignore
        # Round and scale actions if required
        if _round_actions is not None:
            action = np.round(action * _scale_actions, _round_actions)
        else:
            action *= _scale_actions
        # Some agents (i.e. MPC) can interact with an additional environment
        if self.config.settings.interact_with_env:
            assert (
                self.interaction_env is not None
            ), "Initialized interaction environments could not be found. Call prepare_run first."

            # Perform a step  with the interaction environment and update the normal environment with
            # its observations
            observations, _rewards, dones, info = self.interaction_env.step(action)
            observations = np.array(self.environments.env_method("update", observations, indices=0))
            # Make sure to also reset the environment, if the interaction_env says it's done. For the interaction
            # env this is done inside the vectorizer.
            for idx in range(self.environments.num_envs):
                if dones[idx]:
                    info[idx]["terminal_observation"] = observations
                    observations[idx] = self._reset_env_interaction(observations)
        else:
            observations, _rewards, dones, info = self.environments.step(action)
        return observations, dones

    def _reset_envs(self) -> VecEnvObs:
        """Reset the environment before and afterwards when the play and learn function is calling.

        :return: Observations after reset.
        """
        assert self.environments is not None, "Initialized environments could not be found. Call prepare_run first."
        log.debug("Resetting environments.")

        self.environments.seed(self.config.settings.seed)
        if self.config.settings.interact_with_env:
            assert (
                self.interaction_env is not None
            ), "Initialized interaction environments could not be found. Call prepare_run first."
            self.interaction_env.seed(self.config.settings.seed)
            observations = self.interaction_env.reset()
            return self._reset_env_interaction(observations)
        return self.environments.reset()

    def _reset_env_interaction(self, observations: VecEnvObs) -> VecEnvObs:
        """Reset the environments when interaction with another environment is taking place.

        :param Observations: Observations from the interaction env.
        :return: Observations after reset.
        """
        assert self.environments is not None, "Initialized environments could not be found. Call prepare_run first."
        log.debug("Resetting main environment during environment interaction.")

        try:
            observations = np.array(self.environments.env_method("first_update", observations, indices=0))
        except AttributeError as e:
            if "first_update" in str(e):
                observations = self.environments.reset()
            else:
                raise e

        return observations



from __future__ import annotations

from collections import OrderedDict
from copy import deepcopy
from typing import TYPE_CHECKING

import numpy as np
from stable_baselines3.common.vec_env import DummyVecEnv
from stable_baselines3.common.vec_env.util import obs_space_info

if TYPE_CHECKING:
    from typing import Callable

    import gymnasium
    from stable_baselines3.common.vec_env.base_vec_env import (
        VecEnvIndices,
        VecEnvObs,
        VecEnvStepReturn,
    )


class NoVecEnv(DummyVecEnv):
    """
    NoVecEnv is an environment vectorizer which hands the implementation of multithreading off to the environment.
    The environment must specify the attribute "is_multithreaded" and set it to True. NoVecEnv will hand all actions
    it receives over to the environment, even if they were meant for multiple environments. Therefore, the environment
    has to specifically support parallel evaluation of multiple action sets.

    This vectorizer is useful for environments implemented for example in julia, where we do not want to create multiple
    environments and could potentially implement multithreading inside the environment.

    .. note::
        The reset function of the environment should only return a single set of observations. The order in which the
        reset function is called will always be the same. For example, if subenvironments 5, 10 and 15 return done, the
        reset function will be called three times, first for environment 5, then for 10 and finally for 15.

        If the environment returns done (even if it is just for a single subenvironment / action set) the reset function
        of the environment will be called to retrieve initial observations for each one of the done returns separately.

    All other functionality is directly derived from DummyVecEnv. Since there is only a single

    See also: :py:class:`stable_baselines3.common.vec_env.DummyVecEnv`

    :param env_fns: A list of functions that will create the environments
        (each callable returns a `gymnasium.Env` instance when called).
    """

    def __init__(self, env_fns: list[Callable]) -> None:
        super().__init__([env_fns[0]])

        # Correct the number of environments and re-initialize all values which depend on the number of environments.
        self.num_envs = len(env_fns)

        self.keys, shapes, dtypes = obs_space_info(self.envs[0].observation_space)

        self.buf_obs = OrderedDict(
            [(k, np.zeros((self.num_envs, *tuple(shapes[k])), dtype=dtypes[k])) for k in self.keys]
        )
        self.buf_dones = np.zeros((self.num_envs,), dtype=bool)
        self.buf_rews = np.zeros((self.num_envs,), dtype=np.float32)
        self.buf_infos = [{} for _ in range(self.num_envs)]

        self.actions: np.ndarray

        # Check if the environment itself is multithreaded and raise an error if it is not
        if not (hasattr(self.envs[0], "is_multithreaded") and self.envs[0].is_multithreaded is True):
            raise ValueError(
                "The given environment cannot be used with NoVecEnv because it does not specify the attribute "
                "is_multithreaded (or the attribute is not set to True)."
            )

    def step_wait(self) -> VecEnvStepReturn:
        """Store observations and reset environments

        :return: Tuple with stepping result sequences (observations, rewards, dones, infos)
        """
        assert self.actions is not None, "Stepping the environment is only possible when actions are set."

        # Re-initialize the observation buffers (necessary because the number of action sets is not known beforehand).
        obs, self.buf_rews, _terminated, truncated, self.buf_infos = self.envs[0].step(self.actions)  # type: ignore

        for idx in range(self.num_envs):
            self.buf_dones[idx] = _terminated[idx]  # type: ignore
            self.buf_infos[idx]["TimeLimit.truncated"] = truncated[idx] and not _terminated[idx]  # type: ignore

            if self.buf_dones[idx]:
                self.buf_infos[idx]["terminal_observation"] = obs[idx]
                obs[idx], self.reset_infos[0] = self.envs[0].reset()
            self._save_obs(idx, obs[idx])

        # The type of the return value is currently not correct because stablesbaslines3 v2.2.1 has not completely
        # migrated gymnasium into the project but ETA-X still needs five return parameters.
        return (  # type: ignore
            self._obs_from_buf(),
            np.copy(self.buf_rews),
            np.copy(self.buf_dones),
            np.copy(truncated),
            deepcopy(self.buf_infos),
        )

    def reset(self) -> VecEnvObs:
        """Reset all sub environments and return their observations.

        :return: Observations from all sub environments.
        """
        maybe_options = {"options": self._options[0]} if self._options[0] else {}
        for env_idx in range(self.num_envs):
            obs, _ = self.envs[0].reset(seed=self._seeds[env_idx], **maybe_options)
            self._save_obs(env_idx, obs)
        # Seeds and options are only used once
        self._reset_seeds()
        self._reset_options()
        return self._obs_from_buf()

    def _get_target_envs(self, indices: VecEnvIndices) -> list[gymnasium.Env]:
        """Return the 0 target environment (because there can only ever be one...)

        :param indices: Indices of the environments. Values don't not really matter here, only length is important.
        :return: List of environments to target.
        """
        indices = self._get_indices(indices)
        return [self.envs[0] for i in indices]



from __future__ import annotations

from typing import TYPE_CHECKING

import torch as th

if TYPE_CHECKING:
    from collections.abc import Iterable, Sequence


class Split1d(th.nn.ModuleList):
    """Split1d defines a pytorch module which splits the 1D input tensor into multiple parts and passes each
    of the parts through a separate network. After the pass through the network, the output from all networks
    is joined together. Thus, Split1d will return a 1d observation vector.

    When configuring the network architecture, it is important to ensure that the output of all networks is 1D.
    Use torch.nn.Flatten to flatten the output of networks where the output is not one dimensional.

    Use the parameters 'sizes' and 'net_arch' to determine how many of the input features should be passed through
    which network. Each value in sizes must have a correstponding value in net_arch. For the following examples, let's
    assume that 'in_features' is 15. If 'sizes' is [3, 10, None], a valid configuration for
    net_arch could be [th.nn.Linear(out_features=10), th.nn.Conv1d(out_channels:2), th.nn.Linear(out_features=2)].
    The last value of 'sizes' will automatically be calculated to be 2 (15 - 3 - 10 = 2). With this, 3 values would
    be passed to the first *Linear* layer, 10 values would be passed to the "Conv1d" layer and the final 2 values would
    be passed to the third layer in net_arch (which is the *Linear* layer with 2 output features).

    If you would like to use dictionaries to configure the net_arch, you can use the function
    :py:func:`eta_utility.eta_x.common.common.deserialize_net_arch` to create the torch network architecture.

    :param in_features: Number of input features for the Module
    :param sizes: List of sizes for splitting the input features. This list can contain the value "None" once. If the
        list contains None, this will be evaluated to contain all remaining input features.
    :param net_arch: List of torch.nn Modules. Each value of this list corresponds to one value of the 'sizes' list.
    """

    def __init__(self, in_features: int, sizes: Sequence[None | int], net_arch: Sequence[th.nn.Module]):
        super().__init__()

        self.sizes = self.get_full_sizes(in_features, sizes)
        self.in_features = in_features

        # Check that the number of extractor architectures is equal to the  number of sizes specified.
        if len(net_arch) != len(self.sizes):
            raise ValueError(
                f"There must be one extractor architecture (there are {len(net_arch)}) "
                f"for each split in the data (there are {len(self.sizes)})."
            )

        for net in net_arch:
            self.append(net)

    def extra_repr(self) -> str:
        """Add info about the module to its torch representation.

        :return: String representation of the object.
        """
        return f"in_features={self.in_features}, sizes={self.sizes}"

    def forward(self, tensor: th.Tensor) -> th.Tensor:
        """Perform a forward pass through the layer.

        :param tensor: Input tensor
        :return: Output tensor
        """
        if tensor.shape[1] != self.in_features:
            raise ValueError(
                f"The tensor is shorter ({len(tensor)}) than the number of elements specified "
                f"for the split process ({self.in_features})"
            )

        tensors = th.split(tensor, self.sizes, dim=1)
        outputs = [self[item](tensor) for item, tensor in enumerate(tensors)]

        return th.cat(outputs, dim=1)

    @staticmethod
    def get_full_sizes(in_features: int, sizes: Iterable[None | int]) -> list[int]:
        """Use in_features and the sizes list to determine the missing value in 'sizes' in case 'sizes' contains a
        None value (see class description for more information on how a None value in 'sizes' is interpreted.

        :param in_features: Number of input features for the Module.
        :param sizes: List of sizes for splitting the input features. This list can contain the value "None" once. If
            the list contains None, this will be evaluated to contain all remaining input features.
        :return: List of sizes without the missing value.
        """
        # Check if the sizes list contains None and sum all elements that are not None.
        nones = 0
        none_idx: int = 0
        _sum = 0
        for idx, s in enumerate(sizes):
            if s is None:
                none_idx = idx
                nones += 1
            else:
                _sum += s

        if nones > 1:
            raise ValueError(
                "Please only specify None once in the configuration for the split process. None is where "
                "all remaining elements will be processed."
            )
        if nones == 1:
            # mypy does not correctly understand how we are removing None values.
            _sizes: list[int] = list(sizes)  # type: ignore
            _sizes[none_idx] = in_features - _sum
        else:
            if _sum != in_features:
                raise ValueError(
                    f"If None is not specified in the split process configuration, the sum of elements "
                    f"specified in 'sizes' ({_sum}) must be equal to in_features ({in_features})."
                )
            # mypy does not correctly understand how we are removing None values.
            _sizes = list(sizes)  # type: ignore

        return _sizes


class Fold1d(th.nn.Module):
    """Fold a 1D tensor to create a multi-dimensional tensor. The parameter 'out_channels' determines, how many
    dimensions the output tensor will have.

    :param out_channels: Number of dimensions of the output tensor.
    """

    def __init__(self, out_channels: int):
        super().__init__()
        self.out_channels = out_channels

    def extra_repr(self) -> str:
        """Add info about the module to its torch representation.

        :return: String representation of the object.
        """
        return f"out_channels: {self.out_channels}"

    def forward(self, tensor: th.Tensor) -> th.Tensor:
        """Perform a forward pass through the layer.

        :param tensor: Input tensor
        :return: Output tensor
        """
        return th.reshape(tensor, [-1, self.out_channels, tensor.shape[1] // self.out_channels])



from datetime import datetime, timedelta

import pandas as pd

from eta_utility.connectors.cumulocity import CumulocityConnection
from eta_utility.connectors.node import NodeCumulocity

# Params for example Cumulocity-Connection
URL = "eta-iot-cumulocity.ptw.maschinenbau.tu-darmstadt.de"
DEVICE_ID = "2496822328"
MEASUREMENT = "Example-Measurement"
FRAGMENT = "Power"
USER = ""  # enter username here
PASSWORD = ""  # enter your password here
TENANT = "edge"

cumu_node = NodeCumulocity(
    "Cumulocity-Test-Node",
    URL,
    "cumulocity",
    device_id=DEVICE_ID,
    measurement=MEASUREMENT,
    fragment=FRAGMENT,
)

cumu_node2 = NodeCumulocity(
    "Cumulocity-Test-Node2",
    URL,
    "cumulocity",
    device_id=DEVICE_ID,
    measurement=MEASUREMENT,
    fragment=FRAGMENT,
)

### Create Connection ###
# 1. Example: Create connection via connection class
conn = CumulocityConnection(url=URL, usr=USER, pwd=PASSWORD, tenant=TENANT)

# 2. Example: Create connection by Node
conn2 = CumulocityConnection.from_node([cumu_node, cumu_node2], usr=USER, pwd=PASSWORD, tenant=TENANT)
# this way the nodes are directly saved in conn.selected_nodes

### Write Data ###
# Create Example data
t = datetime.now()
index = [t, t + timedelta(minutes=1), t + timedelta(minutes=2)]
data = pd.DataFrame(data=[[1], [2], [3]], columns=["P"], index=index)

# Write data
# to write data you need to provide the additional parameters measurement_type and unit and pass the data as pd.Series
# you can either provide the upload nodes manually
conn.write(values=data["P"], measurement_type="Power", unit="W", nodes=cumu_node)

### Read Data ###
# to read data you need to provide the nodes to read from and a timespan
# again you can either provide the download nodes manually
print(conn.read_series(from_time=t, to_time=t + timedelta(minutes=5), nodes=cumu_node))  # noqa

### Create Device ###
# to create a device, additionally to your login information, you need to provide a device name
DEVICE_NAME = ""
# CumulocityConnection.create_device(url=URL,
# username=USER, password=PASSWORD, tenant=TENANT, device_name=DEVICE_NAME)



"""Initiate live connections that automate certain tasks associated with the creation of such connections."""

from __future__ import annotations

import pathlib
import sys
import time
from collections.abc import Mapping, Sequence
from concurrent.futures import TimeoutError as ConTimeoutError
from contextlib import AbstractContextManager
from datetime import timedelta
from typing import TYPE_CHECKING

import numpy as np

from eta_utility import get_logger, json_import
from eta_utility.connectors import name_map_from_node_sequence
from eta_utility.connectors.base_classes import Connection
from eta_utility.connectors.node import Node

if TYPE_CHECKING:
    import types
    from typing import Any

    from typing_extensions import Self

    from eta_utility.type_hints import Path, TimeStep

log = get_logger("connectors.live")


class LiveConnect(AbstractContextManager):
    """Connect to a system/resource and enable system activation, deactivation and configuring controller setpoints.

    The class can be initialized directly by supplying all required arguments or from a JSON configuration file.

    If initializing with the from_json classmethod, the class takes a JSON config file which defines all
    required nodes. The JSON file should have the following format:

        * **system**: Define the system to be controlled. This is the top level block - it may have multiple dictionary
          members which should each specify all the following information. If there are multiple systems, all
          node names will be prefixed with the system name. For example, if the system name is 'chp' and the node
          name is 'op_request', the full node name will be 'chp.op_request'.
        * **name**: A name to uniquely identify the system.
        * **servers**: Servers which are responsible for manageing the system. This section should contain a dictionary
          of servers, identified by a unique name. Each server has the following values:

            * **url**: URL of the server (format: netloc:port, without the scheme).
            * **protocol**: Protocol for the connection, for example 'opcua'.
            * **usr**: Username to identify with the server (default: None).
            * **pwd**: Password to log in to the server (default: None).
        * **nodes**: The nodes section contains a list of nodes. Each node is specific to one of the servers specified
          above and has a unique name, by which it can be identified. In detail, each node has the following values:

            * **name**: A name to uniquely identify the node.
            * **server**: Identifier of one of the servers defined above.
            * **dtype**: Data type of the Node on the server (default: float).
            * Other values must be defined depending on the protocol, for example this could be an **opc_id** or
              corresponding values to identify modbus nodes. More detail about this can be found in connectors.Nodes.
        * **set_value**: The set_value is the main value, which should be manipulated. It must have some additional
          information to be processed correctly. This includes:

            * **name**: Name of the node, which is manipulated as the set_value.
            * **min**: Minimum value to be allowed.
            * **max**: Maximum value to be allowed.
            * **threshold**: Activation/Deactivation threshold. The activate action will be executed for values above
              the threshold and the deactivation action will be executed for values below the threshold (see below).
            * **add**: Scale the set_value by adding a set amount.
            * **mult**: Scale the set_value by multiplying by this factor.
        * **activation_indicators**: The values specified in this section are used to determine, whether the system is
          currently active. This is a dictionary of nodes and values, these nodes should be compared against. Each node
          is identified by the name specified above. Each node must have the following values:

            * **compare**: Comparison operation to perform (for example '==' or '<=').
            * **value**: Value to compare against. If the result of all comparisons is True, the system is considered
              to be currently active.
        * **observe**: Values to return after each set operation. This is just a list of node names as specified in
          the nodes section above.
        * **actions**: In the actions section, default values for more complex operations are specified. For example,
          if a system needs to be initialized or deactivated. Each of the actions is a dictionary of node names
          and corresponding values, for example {node1: true, node2: 0.2}. The following actions are defined:

            * **init**: Initialize the system. This is used to make sure that the system is ready to receive control
              values from the connector.
            * **activate**: Activate the system. This is used to set the system to an active state, for example by
              requesting the system to start its operation and choosing and operating mode. This can be used to set
              the system up to receive more detailed control values.
            * **deactivate**: Execute any actions to deactivate the system.
            * **close**: Reset the system. This is used when the connector is closed, to make sure the system is left
              in a safe and clean state.

    If initializing manually, each of the actions and some other information must be specified as parameters to the
    class. This way, the class will not automatically identify different systems and will instead assume that all
    parameters belong to the same system. If system differentiation is required, the node prefixing must be done
    manually. The third option is to initialize from a dictionary using the from_dict function. This works equivalent to
    the from_json function.

    .. warning ::
        Always call the close function after you are done using the connection! This is required even if no nodes
        must be written to reset the system since the connection itself must be closed. Therefore, this class should
        only be called within a try-finally clause. Alternatively the class can be used as a context manager in a with
        statement.

    :param name: Name to uniquely identify the system. The name is also used as the default system prefix.
    :param nodes: Sequence/List of Nodes, which should be used for the actions and for initializing connections.
    :param step_size: Step size (time) for the live connector in time increments.
    :param init: Nodes for initializing the system (see above).
    :param activate: Nodes for activating the system (see above).
    :param deactivate: Nodes for deactivating the system (see above).
    :param close: Nodes to close the connection and reset the system (see above).
    :param observe: Nodes to read from and return after each 'set' operation (see above).
    :param activation_indicators: Nodes that determine, whether the system must be deactivated or activated. This
                                  is used internally in conjunction with the set_value threshold to determine whether
                                  the activate/deactivate methods must be called before setting values.
    :param set_values: Specification of the node which is used for setting the main control value.
    """

    def __init__(
        self,
        nodes: Sequence[Node],
        name: str | None = None,
        step_size: TimeStep = 1,
        max_error_count: int = 10,
        *,
        init: Mapping[str, Any] | None = None,
        activate: Mapping[str, Any] | None = None,
        deactivate: Mapping[str, Any] | None = None,
        close: Mapping[str, Any] | None = None,
        observe: Sequence[str] | None = None,
        activation_indicators: Mapping[str, Any] | None = None,
        set_values: Mapping[str, Mapping[str, Any] | None] | None = None,
    ) -> None:
        #: Name of the system.
        self.name: str | None = name.strip() if name is not None else None
        #: Connection objects to the resources.
        self._connections: dict[str, Connection] = Connection.from_nodes(nodes)

        #: Mapping of all nodes.
        self._nodes: dict[str, Node] = name_map_from_node_sequence(nodes)
        #: Mapping of node names to connections.
        self._connection_map: dict[str, str] = {}
        for node in self._nodes.values():
            if node.url_parsed.netloc is not None:
                self._connection_map[node.name] = node.url_parsed.netloc
            else:
                raise ValueError(f"Node without netloc supplied: {node.name}")
        #: Start time of initialisation.
        self.start_time = time.time()
        #: Step size (time) for the live connector in time increments.
        self.step_size = int(step_size) if not isinstance(step_size, timedelta) else int(step_size.total_seconds())
        #: Current step of the live connector (number of completed steps).
        self.steps_counter: int = 0
        #: Maximum error count when connections in read/write function are aborted.
        self.max_error_count: int = max_error_count
        #: Counts the number of errors when Live Connector logs errors.
        self.error_count: list[int] = [0] * len(self._connections)

        self._init_config_nodes(activate, activation_indicators, close, deactivate, init, observe, set_values)

        if self._init_vals is not None:
            self.write(self._init_vals)

    def _init_config_nodes(
        self,
        activate: Mapping[str, Any] | None,
        activation_indicators: Mapping[str, Any] | None,
        close: Mapping[str, Any] | None,
        deactivate: Mapping[str, Any] | None,
        init: Mapping[str, Any] | None,
        observe: Sequence[str] | None,
        set_values: Mapping[str, Mapping[str, Any] | None] | None,
    ) -> None:
        errors = False
        #: Nodes for initializing the system.
        self._init_vals: dict[str, Any] | None
        self._init_vals, e = self._read_value_mapping(init, context_description="initialization")
        errors |= e
        #: Nodes for closing the connection.
        self._close_vals: dict[str, Any] | None
        self._close_vals, e = self._read_value_mapping(close, context_description="closing connection")
        errors |= e
        #: Nodes for activating the system.
        self._activate_vals: dict[str, Any] | None = {}
        self._activate_vals, e = self._read_value_mapping(
            activate, flatten=False, context_description="system activation"
        )
        errors |= e
        #: Nodes for deactivating the system.
        self._deactivate_vals: dict[str, Any] | None = {}
        self._deactivate_vals, e = self._read_value_mapping(
            deactivate, flatten=False, context_description="system deactivation"
        )
        errors |= e
        #: Nodes to observe.
        self._observe_vals: list[str] | None = [val for val in observe if val is not None] if observe else None
        if self._observe_vals and not set(self._observe_vals).issubset(self._nodes):
            missing_nodes = set(self._observe_vals) - set(self._nodes.keys())
            log.error(
                "Not all observed nodes of the object are configured as nodes. Missing nodes: "
                + ", ".join(missing_nodes)
            )
            errors = True

        #: Configuration for the step set value.
        self._set_values: dict[str, Any] | None = {}
        if set_values is not None:
            nds = set()
            setval = set_values.values() if isinstance(set_values, Mapping) else set_values
            for sys_setval in setval:
                if sys_setval is not None:
                    self._set_values[sys_setval["name"]] = sys_setval
                    nds.add(self._set_values[sys_setval["name"]]["node"])
        if not self._set_values:
            self._set_values = None
        elif not nds.issubset(self._nodes):
            missing_nodes = nds - set(self._nodes.keys())
            log.error(
                "Not all nodes required for setting control values are configured as nodes. Missing nodes: "
                + ", ".join(missing_nodes)
            )
            errors = True
        else:
            for _, set_value in self._set_values.items():
                set_value.setdefault("node", set_value["name"])
                set_value.setdefault("threshold", 0)
                set_value.setdefault("mult", 1)
                set_value.setdefault("add", 0)
                set_value.setdefault("min", None)
                set_value.setdefault("max", None)
        #: Indicator to keep track of to know whether the system must be activated or deactivated. If this is not
        #: set, the user must keep track of when the activation and deactivation functions need to be called. This
        #: may not be necessary for all systems.
        self._activation_indicators: dict[str, Any] | None = {}
        self._activation_indicators, e = self._read_value_mapping(
            activation_indicators,
            flatten=False,
            context_description="checking system activation",
        )
        if e or errors:
            raise KeyError("Not all required nodes are configured.")

    def _read_value_mapping(
        self, values: Mapping[str, Any] | None, flatten: bool = True, *, context_description: str
    ) -> tuple[dict[str, Any] | None, bool]:
        """Read a list of values and deserialize it to a mapping.

        :param values: Values to deserialize.
        :param flatten: Output into a single layer (not separated by system).
        :param context_description: Description to log if mapping fails.
        :return: Tuple of deserialized values and bool indicating an error if true.
        """
        missing_keys = set()
        _vals: dict[str, Any] = {}

        if values is not None:
            for key, sys_val in values.items():
                if isinstance(sys_val, Mapping) and flatten:
                    for k, v in sys_val.items():
                        _vals[k] = v
                elif sys_val is not None:
                    _vals[key] = sys_val

        if not _vals:
            vals = None
        else:
            vals = _vals

            if flatten:
                for key, sys_val in vals.items():
                    if sys_val and key not in self._nodes:
                        missing_keys.add(key)
            else:
                for sys_val in vals.values():
                    if sys_val and not sys_val.keys() <= self._nodes.keys():
                        missing_keys.update(sys_val.keys())
        if missing_keys:
            log.error(
                f"Nodes {', '.join(missing_keys)} required for {context_description} are missing from the node list."
            )

        return vals, bool(missing_keys)

    @classmethod
    def from_json(
        cls,
        files: Path | Sequence[Path],
        usr: str | None = None,
        pwd: str | None = None,
        step_size: TimeStep = 1,
        max_error_count: int = 10,
    ) -> LiveConnect:
        """Initialize the connection directly from JSON configuration files. The file should contain parameters
        as described above. A list of file names can be supplied to enable the creation of larger, combined systems.

        Username and password supplied as keyword arguments will take precedence over information given in
        the config file.

        :param files: Configuration file paths. Accepts a single file or a list of files.
        :param usr: Username for authentication with the resource.
        :param pwd: Password for authentication with the resource.
        :param step_size: Step size (time) for the live connector in time increments.
        :return: LiveConnect instance as specified by the JSON file.
        """
        files = [files] if not isinstance(files, Sequence) else files

        config: dict[str, list[Any]] = {"system": []}
        for file_path in files:
            file = pathlib.Path(file_path) if not isinstance(file_path, pathlib.Path) else file_path
            result = json_import(file)
            if not isinstance(result, dict):
                raise TypeError(f"JSON file {file} must define a dictionary of options.")
            if "system" in result:
                config["system"].extend(result["system"])
            else:
                config["system"].append(result)

        return cls.from_dict(usr=usr, pwd=pwd, step_size=step_size, max_error_count=max_error_count, **config)

    @classmethod
    def from_dict(
        cls,
        usr: str | None = None,
        pwd: str | None = None,
        step_size: TimeStep = 1,
        max_error_count: int = 10,
        **config: Any,
    ) -> LiveConnect:
        """Initialize the connection directly from a config dictionary. The dictionary should contain parameters
        as described above.

        Username and password supplied as keyword arguments will take precedence over information given in
        the config file.

        :param usr: Username for authentication with the resource.
        :param pwd: Password for authentication with the resource.
        :param step_size: Step size (time) for the live connector in time increments.
        :param config: Configuration dictionary.
        :return: LiveConnect instance as specified by the JSON file.
        """

        cls._check_config(config)

        # Initialize the connector objects
        _act_indicators, _actions, nodes, _observe, _set_values = cls._read_config(config, pwd, usr)
        name = config["system"][0]["name"] if len(config["system"]) == 1 else None

        return cls(
            nodes,
            name,
            step_size=step_size,
            max_error_count=max_error_count,
            init=_actions["init"],
            activate=_actions["activate"],
            deactivate=_actions["deactivate"],
            close=_actions["close"],
            observe=_observe,
            activation_indicators=_act_indicators,
            set_values=_set_values,
        )

    @classmethod
    def _check_config(cls, config: Mapping[str, Any]) -> None:
        _req_settings = {"name": {}, "servers": {"url", "protocol"}, "nodes": {"name", "server"}}
        if "system" not in config or not isinstance(config["system"], Sequence) or len(config["system"]) < 1:
            raise KeyError("Could not find a valid 'system' section in the configuration")
        # Check that all required parameters are present in config
        errors = False
        for system in config["system"]:
            for sect in _req_settings:
                if sect not in system:
                    log.error(f"Required parameter '{sect}' not found in configuration.")
                    errors = True
                else:
                    for name in _req_settings[sect]:
                        sec = system[sect].values() if isinstance(system[sect], Mapping) else system[sect]
                        for i in sec:
                            if name not in i:
                                log.error(f"Required parameter '{name}' not found in config section '{sect}'")
                                errors = True
        if errors:
            log.error("Not all required config parameters were found. Exiting.")
            sys.exit(1)

    @classmethod
    def _read_config(
        cls, config: Mapping[str, Any], pwd: str | None, usr: str | None
    ) -> tuple[
        dict[str, dict[str, Any] | None],
        dict[str, dict[str, Any]],
        list[Node],
        list[str],
        dict[str, dict[str, Any] | None],
    ]:
        # Make sure all the required sections exist - they are just none if they are not in the file.
        _nodes: list[Node] = []
        _set_values: dict[str, dict[str, Any] | None] = {}
        _act_indicators: dict[str, dict[str, Any] | None] = {}
        _observe: list[str] = []
        _actions: dict[str, dict[str, Any]] = {"activate": {}, "deactivate": {}, "init": {}, "close": {}}
        for system in config["system"]:
            # Combine config for nodes with server config and add system name to nodes
            for _node in system["nodes"]:
                n = _node.copy()
                server = system["servers"][n.pop("server")]
                if "usr" in server and "pwd" in server:
                    usr = server["usr"]
                    pwd = server["pwd"]

                _nodes.extend(
                    Node.from_dict(
                        {
                            "name": f"{system['name']}.{n.pop('name')}",
                            "url": server["url"],
                            "protocol": server["protocol"],
                            "usr": usr,
                            "pwd": pwd,
                            **n,
                        }
                    )
                )

            # Rename set_value
            if "set_value" in system and system["set_value"] is not None:
                _set_values[system["name"]] = system["set_value"].copy()
                _set_values[system["name"]]["name"] = f"{system['name']}.{system['set_value']['name']}"  # type: ignore
                _set_values[system["name"]]["node"] = (  # type: ignore
                    f"{system['name']}.{system['set_value'].get('node', system['set_value']['name'])}"
                )
            else:
                _set_values[system["name"]] = None

            # Convert activation_indicators
            if "activation_indicators" in system and system["activation_indicators"] is not None:
                act_i = {}
                for name, value in system["activation_indicators"].items():
                    act_i[f"{system['name']}.{name}"] = value
                _act_indicators[system["name"]] = act_i
            else:
                _act_indicators[system["name"]] = None

            # Convert observations
            if "observe" in system and system["observe"] is not None:
                _observe.extend(f"{system['name']}.{name}" for name in system["observe"])

            # Convert actions
            for action in ("init", "close", "activate", "deactivate"):
                if "actions" not in system or action not in system["actions"] or system["actions"][action] is None:
                    _actions[action][system["name"]] = None
                else:
                    act = {}
                    for name, value in system["actions"][action].items():
                        act[f"{system['name']}.{name}"] = value
                    _actions[action][system["name"]] = act
        return _act_indicators, _actions, _nodes, _observe, _set_values

    @property
    def nodes(self) -> Mapping[str, Node]:
        """Mapping of all node objects of the connection."""
        return self._nodes.copy()

    def _activated(self, system: str) -> bool:
        """Current activation status of the system, as determined by the activation_indicator.

        :param system: System for which activation status should be checked.
        """
        check_map = {"==": "__eq__", ">": "__gt__", "<": "__lt__", ">=": "__gte__", "<=": "__lte__"}

        # If activation_indicators are specified, read them and identify the status, otherwise the system is
        # considered to be always active.
        if self._activation_indicators is not None and system not in self._activation_indicators:
            raise KeyError(f"Cannot check status of unknown system {system}")
        if self._activation_indicators is not None and self._activation_indicators[system] is not None:
            values = self.read(*self._activation_indicators[system].keys())
            results = []
            for name, check in self._activation_indicators[system].items():
                try:
                    results.append(getattr(values[name], check_map[check["compare"]])(check["value"]))
                except KeyError as error:
                    raise KeyError(f"Unknown comparison operation {check['compare']}") from error

            activated = sum(results) >= 1
        else:
            activated = True
        return activated

    def step(self, value: Mapping[str, Any]) -> dict[str, Any]:
        """Take the set_value and determine, whether the system must be activated or deactivated. Then set the value
        and finally read and return all values as specified by the 'observe' parameter.

        :param value: Value to use as the control value/set_value.
        :return: Values read from the connection as specified by 'observe' parameter.
        """
        self.steps_counter += 1

        write = {}
        for name, val in value.items():
            n = f"{self.name}.{name}" if "." not in name and self.name is not None else name
            if self._set_values is not None and n in self._set_values:
                system, node = n.split(".")

                if (self._set_values[n]["min"] is not None and self._set_values[n]["min"] > val) or (
                    self._set_values[n]["max"] is not None and val > self._set_values[n]["max"]
                ):
                    raise ValueError(
                        f"Set value for node {n} is out of bounds. Value: {val}; "
                        f"Bounds: [{self._set_values[n]['min']}, {self._set_values[n]['max']}]."
                    )

                # Determine activation status and activate or deactivate the system correspondingly
                if val >= self._set_values[n]["threshold"]:
                    self.activate(system)
                elif val <= self._set_values[n]["threshold"]:
                    self.deactivate(system)

                write[self._set_values[n]["node"]] = (val + self._set_values[n]["add"]) * self._set_values[n]["mult"]
            else:
                write[n] = val

        # Write the scaled control value and return the observed values.
        self.write(write)

        try:
            time.sleep((self.steps_counter * self.step_size) - time.time() + self.start_time)
        except Exception:
            log.error("Step_size between write and read function is too small")

        if self._observe_vals is not None:
            return self.read(*self._observe_vals)
        return {}

    def write(self, nodes: Mapping[str, Any] | Sequence[str], values: Sequence[Any] | None = None) -> None:
        """Write any combination of nodes and values.

        :param nodes: Mapping of Nodes and values or Sequence of node names to write to.
        :param values: If nodes are given as a Sequence, this second parameter determined the values to be written.
                       If nodes are given as a Mapping, this parameter is not required. It defaults to None.
        """
        # Determine, whether nodes and values are given as a mapping in the nodes parameter or separately, using both
        # parameters. In the second case, create the mapping.
        if not isinstance(nodes, Mapping) and values is None:
            raise ValueError("Cannot only give nodes or values, specify both.")
        if not isinstance(nodes, Mapping) and values is not None and len(nodes) != len(values):
            raise ValueError(
                f"Each node must have a corresponding value for writing. "
                f"Nodes and values must be of equal length. Given lengths are"
                f"'nodes': {len(nodes)}, 'values': {len(values)}"
            )
        if not isinstance(nodes, Mapping) and values is not None:
            _nodes = dict(zip(nodes, values))
        elif isinstance(nodes, Mapping):
            _nodes = dict(nodes)

        # Sort nodes to be written by connection
        writes: dict[str, dict[Node, Any]] = {url: {} for url in self._connections}
        for name, value in _nodes.items():
            n = f"{self.name}.{name}" if "." not in name and self.name is not None else name
            writes[self._connection_map[n]][self._nodes[n]] = value

        # Write to all selected nodes for each connection
        for idx, connection in enumerate(self._connections):
            try:
                if writes[connection]:
                    self._connections[connection].write(writes[connection])
                    self.error_count[idx] = 0
            except (ConnectionError, ConTimeoutError) as e:
                if self.error_count[idx] < self.max_error_count:
                    self.error_count[idx] += 1
                    log.error(e)
                else:
                    raise

    def read(self, *nodes: str) -> dict[str, Any]:
        """Take a list of nodes and return their names and most recent values.

        :param nodes: One or more nodes to read.
        :return: Dictionary of the most current node values.
        """
        # Sort nodes to be read by connection
        node_readings: dict[str, list[Node]] = {url: [] for url in self._connections}
        _error = False
        for name in nodes:
            n = f"{self.name}.{name}" if "." not in name and self.name is not None else name
            try:
                node_readings[self._connection_map[n]].append(self._nodes[n])
            except KeyError as e:
                new_e = KeyError(f"{e.args[0]} not found in node list.")
                log.error(new_e)
                _error = True
        if _error:
            raise KeyError("Check log. One or more nodes not found in node list.")
        # Read from all selected nodes for each connection
        result = {}
        for idx, connection in enumerate(self._connections):
            try:
                if node_readings[connection]:
                    result.update(self._connections[connection].read(node_readings[connection]).iloc[0].to_dict())
                    self.error_count[idx] = 0
            except (ConnectionError, ConTimeoutError) as e:
                if self.error_count[idx] < self.max_error_count:
                    self.error_count[idx] += 1
                    result.update({name.name: np.nan for name in node_readings[connection]})
                    log.error(e)
                else:
                    raise

        return result

    def activate(self, system: str | None = None) -> None:
        """Take the list of nodes to activate and set them to the correct values to activate the system.

        :param system: System for which should be activated (default: self.name).
        """
        _system = self.name if system is None else system

        if (
            _system is not None
            and not self._activated(_system)
            and self._activate_vals is not None
            and self._activate_vals[_system] is not None
        ):
            self.write(self._activate_vals[_system])

    def deactivate(self, system: str | None = None) -> None:
        """Take the list of nodes to deactivate and set them to the correct values to deactivate the system.

        :param system: System for which should be activated (default: self.name).
        """
        _system = self.name if system is None else system

        if (
            _system is not None
            and self._activated(_system)
            and self._deactivate_vals is not None
            and self._deactivate_vals[_system] is not None
        ):
            self.write(self._deactivate_vals[_system])

    def close(self) -> None:
        """Reset the system and close the connection."""
        if self._close_vals is not None:
            self.write(self._close_vals)

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: types.TracebackType | None
    ) -> None:
        self.close()



from __future__ import annotations

from abc import ABC, abstractmethod


class BaseSchedule(ABC):
    """BaseSchedule provides basic functionality for the implementation of new schedules. Each schedule should
    define a value function.
    """

    @abstractmethod
    def value(self, progress_remaining: float) -> float:
        """Calculate the value of the learning rate based on the remaining progess.

        :param progress_remaining: Remaing progress, which is calculcated in the base class: 1 (start), 0 (end).
        :return: Output value.
        """
        raise NotImplementedError("You can only instantiate subclasses of BaseSchedule.")

    def __call__(self, progress_remaining: float) -> float:
        """Take the current progress remaining and return the result of self.value."""
        return self.value(progress_remaining)

    def __repr__(self) -> str:
        """Representation of the Schedule

        :return: String representation.
        """
        return f"{self.__class__.__name__}({', '.join([f'{name}={value}' for name, value in self.__dict__.items()])})"


class LinearSchedule(BaseSchedule):
    """
    Linear interpolation schedule adjusts the learning rate between initial_p and final_p.
    The value is calculated based on the remaining progress, which is between 1 (start) and 0 (end).

    :param initial_p: Initial output value.
    :param final_p: Final output value.
    """

    def __init__(self, initial_p: float, final_p: float):
        self.initial_p = initial_p
        self.final_p = final_p

    def value(self, progress_remaining: float) -> float:
        """Calculate the value of the learning rate based on the remaining progess.

        :param progress_remaining: Remaing progress, which is calculcated in the base class: 1 (start), 0 (end).
        :return: Output value.
        """
        return self.final_p + progress_remaining * (self.initial_p - self.final_p)



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from eta_utility.eta_x.agents import RuleBased

if TYPE_CHECKING:
    from typing import Any

    from stable_baselines3.common.base_class import BasePolicy
    from stable_baselines3.common.vec_env import VecEnv


class DampedOscillatorControl(RuleBased):
    """
    Simple controller for input signal of damped oscillator model.

    :param policy: Agent policy. Parameter is not used in this agent and can be set to NoPolicy.
    :param env: Environment to be controlled.
    :param verbose: Logging verbosity.
    :param p: Proportional factor for the PID controller
    :param i: Integral factor for the PID controller
    :param d: Derivative factor for the PID controller
    :param kwargs: Additional arguments as specified in stable_baselins3.commom.base_class.
    """

    def __init__(
        self, policy: type[BasePolicy], env: VecEnv, verbose: int = 1, *, p: float, i: float, d: float, **kwargs: Any
    ):
        super().__init__(policy=policy, env=env, verbose=verbose, **kwargs)

        #: Proportional factor for the PID controller.
        self.p = p
        #: Integral factor for the PID controller.
        self.i = i
        #: Derivative factor for the PID controller.
        self.d = d
        #: Integral error value.
        self.integral = 0

    def control_rules(self, observation: np.ndarray) -> np.ndarray:
        """
        Controller of the model. This implements a simple PID controller

        :param observation: Observation from the environment.
        :returns: Resulting action from the PID controller.
        """
        s, v, a = observation
        self.integral += s

        return np.array([-(self.p * s + self.i * self.integral + self.d * v)])



from __future__ import annotations

import pathlib
from datetime import datetime
from functools import partial
from typing import TYPE_CHECKING

import numpy as np

from eta_utility import get_logger
from eta_utility.eta_x.envs import BaseEnv
from eta_utility.util_julia import check_julia_package

if check_julia_package():
    from julia import Main as Jl

    from eta_utility.util_julia import import_jl_file

if TYPE_CHECKING:
    from types import ModuleType
    from typing import Any, Callable

    from eta_utility.eta_x import ConfigOptRun
    from eta_utility.type_hints import StepResult, TimeStep


Jl.eval("using PyCall")
jl_setattribute = Jl.eval("pyfunction(setfield!, PyAny, Symbol, PyAny)")

log = get_logger("eta_x.envs")


class JuliaEnv(BaseEnv):
    """
    TODO: UPDATE DOCUMENTATION!
    Abstract environment definition, providing some basic functionality for concrete environments to use.
    The class implements and adapts functions from gymnasium.Env. It provides additional functionality as required by
    the ETA-X framework and should be used as the starting point for new environments.

    The initialization of this superclass performs many of the necessary tasks, required to specify a concrete
    environment. Read the documentation carefully to understand, how new environments can be developed, building on
    this starting point.

    There are some attributes that must be set and some methods that must be implemented to satisfy the interface. This
    is required to create concrete environments.
    The required attributes are:

        - **version**: Version number of the environment.
        - **description**: Short description string of the environment.
        - **action_space**: The action space of the environment (see also gymnasium.spaces for options).
        - **observation_space**: The observation space of the environment (see also gymnasium.spaces for options).

    The gymnasium interface requires the following methods for the environment to work correctly within the framework.
    Consult the documentation of each method for more detail.

        - **step()**
        - **reset()**
        - **close()**

    :param env_id: Identification for the environment, useful when creating multiple environments.
    :param config_run: Configuration of the optimization run.
    :param verbose: Verbosity to use for logging.
    :param callback: callback which should be called after each episode.
    :param scenario_time_begin: Beginning time of the scenario.
    :param scenario_time_end: Ending time of the scenario.
    :param episode_duration: Duration of the episode in seconds.
    :param sampling_time: Duration of a single time sample / time step in seconds.
    :param render_mode: Renders the environments to help visualise what the agent see, examples
        modes are "human", "rgb_array", "ansi" for text.
    :param kwargs: Other keyword arguments (for subclasses).
    """

    version = "1.0"
    description = "This environment uses a julia file to perform its functions."

    def __init__(
        self,
        env_id: int,
        config_run: ConfigOptRun,
        verbose: int = 2,
        callback: Callable | None = None,
        *,
        scenario_time_begin: datetime | str,
        scenario_time_end: datetime | str,
        episode_duration: TimeStep | str,
        sampling_time: TimeStep | str,
        julia_env_file: pathlib.Path | str,
        render_mode: str | None = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            env_id=env_id,
            config_run=config_run,
            verbose=verbose,
            callback=callback,
            scenario_time_begin=scenario_time_begin,
            scenario_time_end=scenario_time_end,
            episode_duration=episode_duration,
            sampling_time=sampling_time,
            render_mode=render_mode,
        )
        # Set arguments as instance parameters.
        for key, value in kwargs.items():
            setattr(self, key, value)

        #: Julia file name.
        julia_env_file = julia_env_file if isinstance(julia_env_file, pathlib.Path) else pathlib.Path(julia_env_file)

        #: Root Path to the julia file.
        self.julia_env_path: pathlib.Path = (
            julia_env_file if julia_env_file.is_absolute() else config_run.path_root / julia_env_file
        )

        #: Imported Julia file as a module (written in julia) for further initialization of the environment.
        self.__jl: ModuleType = import_jl_file(self.julia_env_path)

        # Make sure that all required functions are implemented in julia.
        for func in ("Environment", "step!", "reset!", "close!", "render", "first_update!", "update!"):
            if not hasattr(self.__jl, func):
                raise NotImplementedError(
                    f"Implementation of abstract method {func} missing from julia implementation of JuliaEnv."
                )

        #: Initialized julia environment (written in julia).
        self._jlenv = self.__jl.Environment(self)

    def first_update(self, observations: np.ndarray) -> np.ndarray:
        """Perform the first update and set values in simulation model to the observed values.

        :param observations: Observations of another environment.
        :return: Full array of observations.
        """
        return self.__jl.first_update_b(observations)

    def update(self, observations: np.ndarray) -> np.ndarray:
        """Update the optimization model with observations from another environment.

        :param observations: Observations from another environment
        :return: Full array of current observations
        """
        return self.__jl.update_b(observations)

    def step(self, action: np.ndarray) -> StepResult:
        """Perform one time step and return its results. This is called for every event or for every time step during
        the simulation/optimization run. It should utilize the actions as supplied by the agent to determine the new
        state of the environment. The method must return a five-tuple of observations, rewards, terminated, truncated,
        info.

        .. note ::
            Do not forget to increment n_steps and n_steps_longtime.

        :param action: Actions taken by the agent.
        :return: The return value represents the state of the environment after the step was performed.

            * **observations**: A numpy array with new observation values as defined by the observation space.
              Observations is a np.array() (numpy array) with floating point or integer values.
            * **reward**: The value of the reward function. This is just one floating point value.
            * **terminated**: Boolean value specifying whether an episode has been completed. If this is set to true,
              the reset function will automatically be called by the agent or by eta_i.
            * **truncated**: Boolean, whether the truncation condition outside the scope is satisfied.
              Typically, this is a timelimit, but could also be used to indicate an agent physically going out of
              bounds. Can be used to end the episode prematurely before a terminal state is reached. If true, the user
              needs to call the `reset` function.
            * **info**: Provide some additional info about the state of the environment. The contents of this may
              be used for logging purposes in the future but typically do not currently serve a purpose.

        """
        self._actions_valid(action)
        self.n_steps += 1

        observations, reward, terminated, truncated, info = self.__jl.step_b(self._jlenv, action)
        self.state_log.append(observations)

        # Render the environment at each step
        if self.render_mode is not None:
            self.render()

        return observations, reward, terminated, truncated, info

    def _reduce_state_log(self) -> list[dict[str, float]]:
        """Removes unwanted parameters from state_log before storing in state_log_longtime

        :return: The return value is a list of dictionaries, where the parameters that
                 should not be stored were removed
        """
        return self.state_log

    def reset(
        self,
        *,
        seed: int | None = None,
        options: dict[str, Any] | None = None,
    ) -> tuple[np.ndarray, dict[str, Any]]:
        """Resets the environment to an initial internal state, returning an initial observation and info.

        This method generates a new starting state often with some randomness to ensure that the agent explores the
        state space and learns a generalised policy about the environment. This randomness can be controlled
        with the ``seed`` parameter otherwise if the environment already has a random number generator and
        :meth:`reset` is called with ``seed=None``, the RNG is not reset. When using the environment in conjunction with
        *stable_baselines3*, the vectorized environment will take care of seeding your custom environment automatically.

        For Custom environments, the first line of :meth:`reset` should be ``super().reset(seed=seed)`` which implements
        the seeding correctly.

        .. note ::
            Don't forget to store and reset the episode_timer.

        :param seed: The seed that is used to initialize the environment's PRNG (`np_random`).
                If the environment does not already have a PRNG and ``seed=None`` (the default option) is passed,
                a seed will be chosen from some source of entropy (e.g. timestamp or /dev/urandom).
                However, if the environment already has a PRNG and ``seed=None`` is passed, the PRNG will *not* be
                reset. If you pass an integer, the PRNG will be reset even if it already exists. (default: None)
        :param options: Additional information to specify how the environment is reset (optional,
                depending on the specific environment) (default: None)

        :return: Tuple of observation and info. The observation of the initial state will be an element of
                :attr:`observation_space` (typically a numpy array) and is analogous to the observation returned by
                :meth:`step`. Info is a dictionary containing auxiliary information complementing ``observation``. It
                should be analogous to the ``info`` returned by :meth:`step`.
        """
        super().reset(seed=seed, options=options)

        # Render the environment when calling the reset function
        if self.render_mode is not None:
            self.render()

        return self.__jl.reset_b(self._jlenv, seed, options)

    def close(self) -> None:
        """Close the environment. This should always be called when an entire run is finished. It should be used to
        close any resources (i.e. simulation models) used by the environment.
        """
        return self.__jl.close_b(self._jlenv)

    def render(self, **kwargs: Any) -> None:
        """Render the environment

        The set of supported modes varies per environment. Some environments do not support rendering at
        all. By convention in Farama *gymnasium*, if mode is:

            * human: render to the current display or terminal and return nothing. Usually for human consumption.
            * rgb_array: Return a numpy.ndarray with shape (x, y, 3), representing RGB values for an x-by-y pixel image,
              suitable for turning into a video.
            * ansi: Return a string (str) or StringIO.StringIO containing a terminal-style text representation.
              The text can include newlines and ANSI escape sequences (e.g. for colors).

        """
        self.__jl.render(self._jlenv, **kwargs)

    def __getattr__(self, name: str) -> Any:
        # Return the item if it is set on the python object
        try:
            return super().__getattribute__(name)
        except AttributeError:
            pass

        # If the item isn't set on the python object, check whether _jlenv exists and has the item
        if "_jlenv" in self.__dict__:
            try:
                return getattr(self._jlenv, name)
            except Exception:
                pass

            try:
                return partial(getattr(self.__jl, name), self._jlenv)
            except Exception:
                pass

        raise AttributeError(f"Could not get {name} from python or julia environment.")

    def __setattr__(self, name: str, value: Any) -> None:
        # Try to set on _jlenv
        if "_jlenv" in self.__dict__ and hasattr(self._jlenv, name):
            try:
                jl_setattribute(self._jlenv, name, value)
            except BaseException as e:
                raise AttributeError(f"Could not set {name} on julia environment: {e}") from e

        # Otherwise set on the python environment.
        super().__setattr__(name, value)



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from eta_utility.eta_x.agents import RuleBased
from eta_utility.eta_x.common import is_vectorized_env

if TYPE_CHECKING:
    from typing import Any

    from stable_baselines3.common.base_class import BasePolicy
    from stable_baselines3.common.vec_env import VecEnv


class PendulumControl(RuleBased):
    def __init__(self, policy: type[BasePolicy], env: VecEnv, verbose: int = 1, **kwargs: Any) -> None:
        """
        Simple controller to test environments.

        :param policy: The policy model to use (not relevant here).
        :param env: The environment to learn from.
        :param verbose: Logging verbosity.
        :param kwargs: Additional arguments as specified in
            :py:class:`stable_baselins3.commom.base_class.BaseAlgorithm`.
        """

        super().__init__(policy=policy, env=env, verbose=verbose, **kwargs)
        assert self.action_space is not None, "action_space not initialized correctly."
        # set initial state
        self.initial_state = np.zeros(self.action_space.shape)  # type: ignore

    def control_rules(self, observation: np.ndarray) -> np.ndarray:
        """This function is abstract and should be used to implement control rules which determine actions from
        the received observations.

        :param observation: Observations as provided by a single, non vectorized environment.
        :return: Action values, as determined by the control rules.
        """
        # Handle vectorized environments
        # correct type ensured by is_vectorized_env
        if is_vectorized_env(self.env) and self.env.num_envs > 1:  # type: ignore
            raise ValueError("The PendulumController can only work on a single environment at once")

        # Calculate actions according to the observations (manually extracting observation values
        cos_th = observation[0]
        # observation[1] (sin_th) not needed here
        th_dot = observation[2]

        # Control rules, can you find a better one? :)
        torque = abs(cos_th) * th_dot
        if cos_th < abs(0.3):
            torque *= 0.3
        else:
            torque *= -1.5

        action: np.ndarray = np.fromiter([torque], dtype=np.floating)

        return action



from .util import (
    LOG_DEBUG as LOG_DEBUG,
    LOG_ERROR as LOG_ERROR,
    LOG_INFO as LOG_INFO,
    LOG_WARNING as LOG_WARNING,
    KeyCertPair as KeyCertPair,
    PEMKeyCertPair as PEMKeyCertPair,
    SelfsignedKeyCertPair as SelfsignedKeyCertPair,
    Suppressor as Suppressor,
    deep_mapping_update as deep_mapping_update,
    dict_get_any as dict_get_any,
    dict_pop_any as dict_pop_any,
    dict_search as dict_search,
    ensure_timezone as ensure_timezone,
    get_logger as get_logger,
    json_import as json_import,
    log_add_filehandler as log_add_filehandler,
    url_parse as url_parse,
)
from .util_julia import (
    install_julia as install_julia,
    update_agent as update_agent,
)



"""The FMUSimulator class enables easy simulation of FMU files."""

from __future__ import annotations

import itertools as it
import shutil
from collections.abc import Mapping
from datetime import timedelta
from typing import TYPE_CHECKING

import numpy as np
from fmpy import extract, read_model_description
from fmpy.fmi2 import FMU2Model, FMU2Slave
from fmpy.sundials import CVodeSolver
from fmpy.util import compile_platform_binary

from eta_utility import get_logger

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import Any

    from fmpy.model_description import ModelDescription

    from eta_utility.type_hints import Number, Path, TimeStep

log = get_logger("simulators.fmu")


class FMUSimulator:
    """FMU simulator object.

    :param _id: FMU instance ID.
    :param fmu_path: Path to the FMU file.
    :param start_time: Simulation start time in seconds.
    :param float stop_time: Simulation stop time in seconds.
    :param step_size: Simulation step size in seconds.
    :param names_inputs: List of input names that correspond to names used in the FMU file (e.g. ['u', 'p']).
                         If the step function is going to be used with lists as input values, this list will be used
                         to translate between the list position and the variable name in the FMU.
    :param names_outputs: List of output names that correspond to names used in the FMU file
                          (e.g. ['y', 'th', 'thdot']). If the step function should return only specific values instead
                          of all results as a dictionary, this parameter can be specified to determine, which parameters
                          should be returned.
    :param init_values: Starting values for parameters that should be pushed to the FMU with names corresponding to
                        variables in the FMU.
    :param str return_type: "dict" or "list". Alter the standard behavior, which is to return lists from the step and
                            get_values functions only if both, "names_inputs" and "names_outputs" are specified.
                            This parameter will force the step and get_values functions to always return either
                            dictionaries or lists.
    """

    def __init__(
        self,
        _id: int,
        fmu_path: Path,
        start_time: TimeStep = 0,
        stop_time: TimeStep = 1,
        step_size: TimeStep = 1,
        names_inputs: Sequence[str] | None = None,
        names_outputs: Sequence[str] | None = None,
        init_values: Mapping[str, float] | None = None,
        *,
        return_type: str | None = None,
    ) -> None:
        #: Path to the FMU model.
        self.fmu_path = fmu_path

        #: Start time for the simulation in time increments.
        self.start_time = start_time.total_seconds() if isinstance(start_time, timedelta) else start_time
        #: Stopping time for the simulation in time increments (only relevant if run in simulation loop).
        self.stop_time = stop_time.total_seconds() if isinstance(stop_time, timedelta) else stop_time
        #: Step size (time) for the simulation in time increments.
        self.step_size = step_size.total_seconds() if isinstance(step_size, timedelta) else step_size

        #: Model description from the FMU (contains variable names, types, references and more).
        self.model_description: ModelDescription = read_model_description(fmu_path)

        #: Variable map from model description. The map specifies the value reference and datatype of a named
        #: variable in the FMU. The structure is {'name': {'ref': <value reference>, 'type': <variable data type>}}.
        self._model_vars: dict[str, dict[str, str]] = {}
        self.__type_map = {"Real": "real", "Boolean": "bool", "Integer": "int", "Enumeration": "enum"}

        for var in self.model_description.modelVariables:
            self._model_vars[var.name] = {"ref": var.valueReference, "type": self.__type_map[var.type]}

        #: Map of input variables which can be used to evaluate an ordered list of input variables. This is typically
        #: not required when working with mappings/dictionaries as step inputs.
        #:
        #: The map contains the following lists:
        #:
        #:     * real: Mask for real variables. This can be used to identify real variables from the complete set of
        #:       input variables (_inputs["refs"], see below) using `itertools.compress`.
        #:     * int: Mask for integer variables. This can be used to identify integer variables from the complete set
        #:       of input variables (_inputs["refs"], see below) using `itertools.compress`.
        #:     * bool: Mask for boolean variables. This can be used to identify boolean variables from the complete set
        #:       of input variables (_inputs["refs"], see below) using `itertools.compress`.
        self._input_map: dict[str, list[bool]] = {"names": [], "real": [], "int": [], "bool": []}

        #: Map of input variable references and their names. The map contains the following lists:
        #:
        #:     * names: List of the named input variables that are accessible in the model.
        #:     * refs: List of all value references to input variables of all types. This is the complete list, which
        #:       can be filtered using itertools.compress (see above).
        #:     * real: List of all value references to input variables of type real.
        #:     * int: List of all value references to input variables of type integer.
        #:     * bool: List of all value references to input variables of type boolean.
        self._inputs: dict[str, list[str]] = {"names": [], "refs": [], "real": [], "int": [], "bool": []}
        refs = []
        names = []
        iterator = names_inputs if names_inputs is not None else self._model_vars.keys()

        for var in iterator:
            if var in self._model_vars:
                refs.append(self._model_vars[var]["ref"])
                names.append(var)
                self._input_map["real"].append(self._model_vars[var]["type"] == "real")
                self._input_map["int"].append(self._model_vars[var]["type"] == "int")
                self._input_map["bool"].append(self._model_vars[var]["type"] == "bool")
            else:
                log.warning(
                    f"Input variable '{var}' couldn't be found in FMU model description. Entry will be ignored."
                )

        self._inputs["names"] = names
        self._inputs["refs"] = refs
        self._inputs["real"] = list(it.compress(refs, self._input_map["real"]))
        self._inputs["int"] = list(it.compress(refs, self._input_map["int"]))
        self._inputs["bool"] = list(it.compress(refs, self._input_map["bool"]))

        #: Map of output variables which can be used to evaluate an ordered list of output variables. This is typically
        #: not required when working with mappings/dictionaries as step outputs.
        #:
        #: The map contains the following lists:
        #:
        #:     * real: Mask for real variables. This can be used to identify real variables from the complete set of
        #:       output variables (_outputs['refs'], see below) using `itertools.compress`.
        #:     * int: Mask for integer variables. This can be used to identify integer variables from the complete set
        #:       of output variables (_outputs['refs'], see below) using `itertools.compress`.
        #:     * bool: Mask for boolean variables. This can be used to identify boolean variables from the complete set
        #:       of output variables (_outputs['refs'], see below) using `itertools.compress`.
        self._output_map: dict[str, list[bool]] = {"names": [], "real": [], "int": [], "bool": []}

        #: Map of output variable references and their names. The map contains the following lists:
        #:
        #:     * names: List of the named output variables that are accessible in the model.
        #:     * refs: List of all value references to output variables of all types. This is the complete list, which
        #:       can be filtered using itertools.compress (see above).
        #:     * real: List of all value references to output variables of type real.
        #:     * int: List of all value references to output variables of type integer.
        #:     * bool: List of all value references to output variables of type boolean.
        self._outputs: dict[str, list[str]] = {"names": [], "refs": [], "real": [], "int": [], "bool": []}
        refs = []
        names = []
        iterator = names_outputs if names_outputs is not None else self._model_vars.keys()

        for var in iterator:
            if var in self._model_vars:
                refs.append(self._model_vars[var]["ref"])
                names.append(var)
                self._output_map["real"].append(self._model_vars[var]["type"] == "real")
                self._output_map["int"].append(self._model_vars[var]["type"] == "int")
                self._output_map["bool"].append(self._model_vars[var]["type"] == "bool")
            else:
                log.warning(
                    f"Output variable '{var}' couldn't be found in FMU model description. Entry will be ignored."
                )
        self._outputs["names"] = names
        self._outputs["refs"] = refs
        self._outputs["real"] = list(it.compress(refs, self._output_map["real"]))
        self._outputs["int"] = list(it.compress(refs, self._output_map["int"]))
        self._outputs["bool"] = list(it.compress(refs, self._output_map["bool"]))

        #: Directory where the FMU will be extracted.
        self._unzipdir: Path = extract(fmu_path)

        try:
            #: Instance of the FMU Slave object.
            self.fmu: FMU2Slave = FMU2Slave(
                guid=self.model_description.guid,
                unzipDirectory=self._unzipdir,
                modelIdentifier=self.model_description.coSimulation.modelIdentifier,
                instanceName="FMUsimulator_" + str(_id),
            )
        except Exception:
            compile_platform_binary(self.fmu_path)
            self.fmu = FMU2Slave(
                guid=self.model_description.guid,
                unzipDirectory=self._unzipdir,
                modelIdentifier=self.model_description.coSimulation.modelIdentifier,
                instanceName="FMUsimulator_" + str(_id),
            )

        # initialize
        self.fmu.instantiate(visible=False, callbacks=None, loggingOn=False)
        self.fmu.setupExperiment(startTime=self.start_time)

        # set init values
        # instead of using the fmpy apply_start_values func from fmpy use the own set_values func to set the values
        # of the simulation variables correctly, reasons are also performance and simulation speed
        init_values = {} if init_values is None else init_values
        self.set_values(init_values)

        self.fmu.enterInitializationMode()
        self.fmu.exitInitializationMode()
        #: Current simulation time.
        self.time = self.start_time

        # Initialize some other parameters used to switch functionality of class methods.
        #: Return dictionaries from the step and get_values functions instead of lists.
        self._return_dict: bool = False
        if return_type is None:
            self._return_dict = names_inputs is None or names_outputs is None
        else:
            self._return_dict = return_type != "list"

    @property
    def input_vars(self) -> list[str]:
        """Ordered list of all available input variable names in the FMU."""
        return self._inputs["names"].copy()

    @property
    def output_vars(self) -> list[str]:
        """Ordered list of all available output variable names in the FMU."""
        return self._outputs["names"].copy()

    def read_values(self, names: Sequence[str] | None = None) -> dict[str | int, Any] | list:
        """Return current values of the simulation without advancing a simulation step or the simulation time.

        :param names: Sequence of values to read from the FMU. If this is None (default), all available values will be
                      read.
        """
        # Find value references and names for the variables that should be read from the FMU
        if names is None:
            refs = self._outputs["refs"]
            vars_ = self._outputs["names"]
        else:
            refs = []
            vars_ = []
            for var in names:
                try:
                    refs.append(self._model_vars[var]["ref"])
                    vars_.append(var)
                except KeyError as e:
                    msg = f"Specified an output value for a variable which is not available in the FMU: {var}"
                    raise KeyError(msg) from e

        # Get values from the FMU and convert to specified output format (dict or list)
        output_values = self.fmu.getReal(refs)
        return dict(zip(vars_, output_values)) if self._return_dict else output_values

    def set_values(self, values: Sequence[Number | bool] | Mapping[str, Number | bool]) -> None:
        """Set values of simulation variables without advancing a simulation step or the simulation time.

        :param values: Values that should be pushed to the FMU. Names of the input_values must correspond
                       to variables in the FMU. If passing as a Sequence, make sure the order corresponds to
                       the order of the input_vars property.
        """

        vals: dict[str, list[Number | bool]] = {"real": [], "int": [], "bool": []}
        refs: dict[str, list[str]] = {"real": [], "int": [], "bool": []}
        if isinstance(values, Mapping):
            for var, val in values.items():
                try:
                    refs[self._model_vars[var]["type"]].append(self._model_vars[var]["ref"])
                    vals[self._model_vars[var]["type"]].append(val)
                except KeyError as e:
                    msg = f"Specified an input value for a variable which is not available in the FMU: {var}"
                    raise KeyError(msg) from e
        else:
            if len(values) != len(self._inputs["refs"]):
                raise AttributeError(
                    f"Length of value list ({len(values)}) must be equal to length of input_vars "
                    f"property ({len(self._inputs['refs'])})"
                )
            refs = {
                "real": self._inputs["real"],
                "int": self._inputs["int"],
                "bool": self._inputs["bool"],
            }

            vals = {
                "real": list(it.compress(values, self._input_map["real"])),
                "int": list(it.compress(values, self._input_map["int"])),
                "bool": list(it.compress(values, self._input_map["bool"])),
            }

        if len(refs["real"]) > 0:
            self.fmu.setReal(refs["real"], vals["real"])
        if len(refs["int"]) > 0:
            self.fmu.setInteger(refs["int"], vals["int"])
        if len(refs["bool"]) > 0:
            self.fmu.setBoolean(refs["bool"], vals["bool"])

    def step(
        self,
        input_values: Sequence[Number | bool] | Mapping[str, Number | bool] | None = None,
        output_names: Sequence[str] | None = None,
        advance_time: bool = True,
        nr_substeps: int | None = None,
    ) -> dict[str | int, Any] | list[Any]:
        """Simulate next time step in the FMU with defined input values and output values.

        :param input_values: Current values that should be pushed to the FMU. Names of the input_values must correspond
                             to variables in the FMU. If passing as a Sequence, make sure the order corresponds to
                             the order of the input_vars property.
        :param advance_time: Decide if the FMUsimulator should add one timestep to the simulation time or not.
                                  This can be deactivated, if you just want to look at the result of a simulation step
                                  beforehand, without actually advancing simulation time.
        :param nr_substeps: if simulation steps are divided into substeps, this value will let the simulator know
                                that no time violation warning is necessary.
        :return: Resulting input and output values from the FMU with the keys named corresponding to the variables
                 in the FMU.
        """
        if input_values is not None:
            self.set_values(input_values)

        # put out warning for time limit violation, if self.time + self.step_size > self.stop_time + full step size
        if self.time + self.step_size > self.stop_time + (int(nr_substeps) if nr_substeps else 1) * self.step_size:
            log.warning(
                f"Simulation time {self.time + self.step_size} s exceeds specified stop time of "
                f"{self.stop_time} s. Proceed with care, simulation may become inaccurate."
            )

        # push input values to the FMU and do one timestep, doStep performs a step of certain size
        self.fmu.doStep(currentCommunicationPoint=self.time, communicationStepSize=self.step_size)

        # advance time
        if advance_time:
            self.time += self.step_size  # advance the time

        return self.read_values(output_names)

    @classmethod
    def simulate(
        cls,
        fmu_path: Path,
        start_time: TimeStep = 0,
        stop_time: TimeStep = 1,
        step_size: TimeStep = 1,
        init_values: Mapping[str, float] | None = None,
    ) -> np.ndarray:
        """Instantiate a simulator with the specified FMU, perform simulation and return results.

        :param fmu_path: Path to the FMU file.
        :param start_time: Simulation start time in seconds.
        :param float stop_time: Simulation stop time in seconds.
        :param step_size: simulation step size in seconds.
        :param init_values: Starting values for parameters that should be pushed to the FMU with names corresponding to
                            variables in the FMU.
        """
        simulator = cls(0, fmu_path, start_time, stop_time, step_size, init_values=init_values)

        dt = np.dtype([(name, float) for name in simulator.read_values()])
        # mypy does not recognize the return type of floor division...
        result = np.rec.array(
            None,
            shape=((simulator.stop_time - simulator.start_time) // simulator.step_size + 1,),  # type: ignore
            dtype=dt,
        )
        assert result.dtype.names is not None, "There must be some output variables specified for the simulator."

        step = 0
        while simulator.time <= simulator.stop_time:
            step_result = simulator.step()
            assert isinstance(step_result, dict), "The simulator needs a dictionary return."
            for name in result.dtype.names:
                result[step][name] = step_result[name]

            step += 1

        return result

    def reset(self, init_values: Mapping[str, float] | None = None) -> None:
        """Reset FMU to specified initial condition.

        :param init_values: Values for initialization.
        """
        self.time = self.start_time
        self.fmu.reset()
        self.fmu.setupExperiment(startTime=self.start_time)

        # set init values
        # instead of using the fmpy apply_start_values func from fmpy use the own set_values func to set the values
        # of the simulation variables correctly, reasons are also performance and simulation speed
        self.set_values(init_values)  # type: ignore

        self.fmu.enterInitializationMode()
        self.fmu.exitInitializationMode()

    def close(self) -> None:
        """Close the FMU and tidy up the unzipped files."""
        self.fmu.terminate()
        self.fmu.freeInstance()
        shutil.rmtree(self._unzipdir)  # clean up unzipped files


class FMU2MESlave(FMU2Model):
    """Helper class for simulation of FMU2 FMUs. This is as wrapper for FMU2Model.
    It can be used to wrap model exchange FMUs such that they can be simulated similar to a co-simulation FMU. This
    is especially helpful for testing model exchange FMUs.

    It exposes an interface that emulates part of the original FMU2Slave class from fmpy.
    """

    # Define some constants that might be needed according to the FMI Standard
    fmi2True: int = 1  # noqa: N815
    fmi2False: int = 0  # noqa: N815

    fmi2OK: int = 0  # noqa: N815
    fmi2Warning: int = 1  # noqa: N815
    fmi2Discard: int = 2  # noqa: N815
    fmi2Error: int = 3  # noqa: N815
    fmi2Fatal: int = 4  # noqa: N815
    fmi2Pending: int = 5  # noqa: N815

    def __init__(self, **kwargs: Any) -> None:
        r"""Initialize the FMU2Slave object. See also the fmyp documentation :py:class:`fmpy.fmi2.FMU2Model`.

        :param Any \**kwargs: Accepts any parameters that fmpy.FMU2Model accepts.
        """

        super().__init__(**kwargs)
        self._model_description: ModelDescription = read_model_description(kwargs["unzipDirectory"])
        self._solver: CVodeSolver
        self._tolerance: float = 0.0
        self._stop_time: float = 0.0
        self._start_time: float = 0.0

    def setupExperiment(  # noqa: N802
        self,
        tolerance: float | None = None,
        startTime: float = 0.0,  # noqa:N803
        stopTime: float | None = None,  # noqa:N803
        **kwargs: Any,
    ) -> int:
        """Experiment setup and storage of required values.

        .. see also::
            fmpy.fmi2.FMU2Model.setupExperiment

        :param tolerance: Solver tolerance, default value is 1e-5.
        :param startTime: Starting time for the experiment.
        :param stopTime: Ending time for the experiment.
        :param kwargs: Other keyword arguments that might be required for FMU2Model.setupExperiment in the future.
        :return: FMI2 return value.
        """

        self._tolerance = 1e-5 if tolerance is None else tolerance
        self._stop_time = 0.0 if stopTime is None else stopTime
        self._start_time = startTime

        kwargs["tolerance"] = self._tolerance
        kwargs["stopTime"] = self._stop_time
        kwargs["startTime"] = self._start_time

        return super().setupExperiment(**kwargs)

    def exitInitializationMode(self, **kwargs: Any) -> int:  # noqa: N802
        """Exit the initialization mode and set up the cvode solver.

        See also: :py:class:`fmpy.fmi2.FMU2Model.exitInitializationMode`

        :param kwargs: Keyword arguments accepted by FMU2Model.exitInitializationMode.
        :return: FMI2 return value.
        """

        ret = super().exitInitializationMode(**kwargs)

        # Collect discrete states from FMU
        self.eventInfo.newDiscreteStatesNeeded = self.fmi2true
        self.eventInfo.terminateSimulation = self.fmi2false

        while (
            self.eventInfo.newDiscreteStatesNeeded == self.fmi2true
            and self.eventInfo.terminateSimulation == self.fmi2false
        ):
            # update discrete states
            self.newDiscreteStates()
        self.enterContinuousTimeMode()

        # Initialize solver
        self._solver = CVodeSolver(
            set_time=self.setTime,
            startTime=self._start_time,
            maxStep=(self._stop_time - self._start_time) / 50.0,
            relativeTolerance=self._tolerance,
            nx=self._model_description.numberOfContinuousStates,
            nz=self._model_description.numberOfEventIndicators,
            get_x=self.getContinuousStates,
            set_x=self.setContinuousStates,
            get_dx=self.getDerivatives,
            get_z=self.getEventIndicators,
        )

        return ret

    def doStep(  # noqa: N802
        self,
        currentCommunicationPoint: float,  # noqa: N803
        communicationStepSize: float,  # noqa: N803
        noSetFMUStatePriorToCurrentPoint: int | None = None,  # noqa: N803
    ) -> int:
        """Perform a simulation step. Advance simulation from *currentCommunicationPoint* by *communicationStepSize*.

        Also refer to the FMI2 Standard documentation.

        :param currentCommunicationPoint: Current time stamp (starting point for simulation step).
        :param communicationStepSize: Time step size.
        :param noSetFMUStatePriorToCurrentPoint: Determine whether a reset before *currentCommunicationPoint* is
            possible. Must be either fmi2True or fmi2False.
        :return: FMU2 return value.
        """
        time = currentCommunicationPoint
        step_size = communicationStepSize

        # Perform a solver step and reset the FMU Model time.
        _, time = self._solver.step(time, time + step_size)
        self.setTime(time)
        # Check for events that might have occurred during the step
        step_event, _ = self.completedIntegratorStep()

        return self.fmi2ok



from eta_utility.util_julia import julia_extensions_available

from .base_env import BaseEnv as BaseEnv
from .base_env_live import BaseEnvLive as BaseEnvLive
from .base_env_mpc import BaseEnvMPC as BaseEnvMPC
from .base_env_sim import BaseEnvSim as BaseEnvSim
from .no_vec_env import NoVecEnv as NoVecEnv
from .state import (
    StateConfig as StateConfig,
    StateVar as StateVar,
)

# Import JuliaEnv if julia is available and ignore errors otherwise.
if julia_extensions_available():
    from .julia_env import JuliaEnv as JuliaEnv



"""
This module provides a read-only REST API connector to the forecast.solar API.

You can obtain an estimate of solar production for a specific location, defined by latitude and longitude,
and a specific plane orientation, defined by declination and azimuth, based on the installed module power.

Supported endpoints include: "Estimate", "Historic", and "Clearsky":

**Estimate Solar Production**
The `estimate` endpoint provides the forecast for today and the upcoming days, depending on the account model.

**Historic Solar Production**
The `historic` endpoint calculates the average solar production for a given day based on historical weather data,
excluding current weather conditions.

**Clear Sky Solar Production**
The `clearsky` endpoint calculates the theoretically possible solar production assuming no cloud cover.

For more information, visit the `forecast.solar API documentation <https://doc.forecast.solar/start>`_.

"""

# Consider Caching for at least 15 minutes and don't exceed rate limit
from __future__ import annotations

import concurrent.futures
import traceback
from collections.abc import Mapping
from datetime import datetime, timedelta
from typing import TYPE_CHECKING

import pandas as pd
import requests
from requests_cache import DO_NOT_CACHE, CachedSession

from eta_utility import get_logger
from eta_utility.connectors.node import NodeForecastSolar
from eta_utility.timeseries import df_resample
from eta_utility.util import round_timestamp

from .base_classes import SeriesConnection, SubscriptionHandler

if TYPE_CHECKING:
    from types import TracebackType
    from typing import Any, ClassVar

    from typing_extensions import Self

    from eta_utility.type_hints import AnyNode, Nodes, TimeStep


log = get_logger("connectors.forecast_solar")


class ForecastSolarConnection(SeriesConnection, protocol="forecast_solar"):
    """
    ForecastSolarConnection is a class to download and upload multiple features from and to the Forecast.Solar database
    as timeseries.

    :param url: URL of the server with scheme (https://).
    :param usr: Not needed for Forecast.Solar.
    :param pwd: Not needed for Forecast.Solar.
    :param api_key: Token for API authentication.
    :param nodes: Nodes to select in connection.
    """

    _baseurl: ClassVar[str] = "https://api.forecast.solar"
    _time_format: ClassVar[str] = "%Y-%m-%dT%H:%M:%SZ"
    _headers: ClassVar[dict[str, str]] = {"Content-Type": "application/json"}

    def __init__(
        self,
        url: str = _baseurl,
        *,
        api_key: str = "None",
        url_params: dict[str, Any] | None = None,
        query_params: dict[str, Any] | None = None,
        nodes: Nodes | None = None,
    ) -> None:
        super().__init__(url, None, None, nodes=nodes)

        #: Url parameters for the forecast.Solar api
        self.url_params: dict[str, Any] | None = url_params
        #: Query parameters for the forecast.Solar api
        self.query_params: dict[str, Any] | None = query_params
        #: Key to use the Forecast.Solar api. If API key is none, only the public functions are usable.
        self._api_key: str = api_key
        #: Cached session to handle the requests
        self._session: CachedSession = CachedSession(
            cache_name="eta_utility/connectors/requests_cache/forecast_solar_cache",
            urls_expire_after={
                "https://api.forecast.solar*": 900,  # 15 minutes
                "*": DO_NOT_CACHE,  # Don't cache other URLs
            },
            allowable_codes=(200, 400, 401, 403),
            use_cache_dir=True,
        )

    @classmethod
    def _from_node(cls, node: AnyNode, **kwargs: Any) -> ForecastSolarConnection:
        """Initialize the connection object from an Forecast.Solar protocol node object

        :param node: Node to initialize from.
        :param kwargs: Keyword arguments for API authentication, where "api_key" is required
        :return: ForecastSolarConnection object.
        """
        api_key = kwargs.get("api_key", "None")

        return super()._from_node(node, api_key=api_key)

    def _read_node(self, node: NodeForecastSolar) -> pd.DataFrame:
        """Download data from the Forecast.Solar Database.

        :param node: Node to read values from.
        :return: pandas.DataFrame containing the data read from the connection.
        """
        url, query_params = node.url, node._query_params

        raw_response = self._raw_request("GET", url, params=query_params, headers=self._headers)
        response = raw_response.json()

        timestamps = pd.to_datetime(list(response["result"].keys()))
        watts = response["result"].values()

        data = pd.DataFrame(
            data=watts,
            index=timestamps.tz_convert(self._local_tz),
            dtype="float64",
        )
        data.index.name = "Time (with timezone)"

        return data

    def _select_data(
        self, results: pd.DataFrame, from_time: pd.Timestamp | None = None, to_time: pd.Timestamp | None = None
    ) -> tuple[pd.DataFrame, pd.Timestamp]:
        """Forecast.solar api returns the data for the whole day. Select data only for the time interval.

        :param nodes: pandas.DataFrame containing the raw data read from the connection.
        :param from_time: Starting time to begin reading (included in output).
        :param to_time: Time to stop reading at (included in output).
        :return: pandas.DataFrame containing the selected data read from the connection and the current timestamp.
        """
        now = pd.Timestamp.now().tz_localize(self._local_tz)

        if isinstance(from_time, pd.Timestamp):
            previous_time = from_time.floor("15T") if self._api_key != "None" else from_time.floor("h")
        else:
            # When from_time is None, no series is selected
            previous_time = now.floor("15T") if self._api_key != "None" else now.floor("h")

        if isinstance(to_time, pd.Timestamp):
            next_time = (
                to_time.floor("15T") + timedelta(minutes=15)
                if self._api_key != "None"
                else to_time.floor("h") + timedelta(minutes=60)
            )
        else:
            # When to_time is None, no series is selected
            next_time = (
                previous_time + timedelta(minutes=15)
                if self._api_key != "None"
                else previous_time + timedelta(minutes=60)
            )

        if previous_time not in results:
            results.loc[previous_time] = 0

        if next_time not in results:
            results.loc[next_time] = 0

        results = results.sort_index()

        return results.loc[previous_time:next_time], now

    def read(self, nodes: Nodes | None = None) -> pd.DataFrame:
        """Return current value from the Forecast.Solar Database.

        :param nodes: List of nodes to read values from.
        :return: pandas.DataFrame containing the data read from the connection.
        """
        nodes = self._validate_nodes(nodes)

        with concurrent.futures.ThreadPoolExecutor() as executor:
            results = executor.map(self._read_node, nodes)

        values = pd.concat(results, axis=1, keys=[n.name for n in nodes], sort=False)

        values, now = self._select_data(values)

        # Insert the current timestamp _now and sort the index column to finish with the linear interpolation method
        values.loc[now] = pd.NA
        values = values.sort_index()

        return pd.DataFrame(values.interpolate(method="linear").loc[now])

    def write(self, values: Mapping[AnyNode, Any]) -> None:
        """
        .. warning::
            Cannot read single values from the Forecast.Solar API. Use read_series instead

        :raises NotImplementedError:
        """
        raise NotImplementedError("Write is not implemented for Forecast.Solar.")

    def subscribe(self, handler: SubscriptionHandler, nodes: Nodes | None = None, interval: TimeStep = 1) -> None:
        """
        .. warning::
            Cannot read single values from the Forecast.Solar API. Use read_series instead

        :raises NotImplementedError:
        """
        raise NotImplementedError("Subscribe is not implemented for Forecast.Solar.")

    def read_series(
        self, from_time: datetime, to_time: datetime, nodes: Nodes | None = None, interval: TimeStep = 1, **kwargs: Any
    ) -> pd.DataFrame:
        """Download timeseries data from the Forecast.Solar Database

        :param nodes: List of nodes to read values from.
        :param from_time: Starting time to begin reading (included in output).
        :param to_time: Time to stop reading at (not included in output).
        :param interval: Interval between time steps. It is interpreted as seconds if given as integer.
        :param kwargs: Other parameters (ignored by this connector).
        :return: Pandas DataFrame containing the data read from the connection.
        """
        _interval = interval if isinstance(interval, timedelta) else timedelta(seconds=interval)

        nodes = self._validate_nodes(nodes)

        from_time = pd.Timestamp(round_timestamp(from_time, _interval.total_seconds())).tz_convert(self._local_tz)
        to_time = pd.Timestamp(round_timestamp(to_time, _interval.total_seconds())).tz_convert(self._local_tz)

        with concurrent.futures.ThreadPoolExecutor() as executor:
            results = executor.map(self._read_node, nodes)

        values = pd.concat(results, axis=1, keys=[n.name for n in nodes], sort=False)

        values, _ = self._select_data(values, from_time, to_time)

        values = df_resample(values, _interval, missing_data="interpolate")

        return values.loc[from_time:to_time]  # type: ignore

    def subscribe_series(
        self,
        handler: SubscriptionHandler,
        req_interval: TimeStep,
        offset: TimeStep | None = None,
        nodes: Nodes | None = None,
        interval: TimeStep = 1,
        data_interval: TimeStep = 1,
        **kwargs: Any,
    ) -> None:
        """
        .. warning::
            Cannot read single values from the Forecast.Solar  API. Use read_series instead

        :raises NotImplementedError:
        """
        raise NotImplementedError("Subscribe series is not implemented for Forecast.Solar.")

    def close_sub(self) -> None:
        """
        .. warning::
            Cannot read single values from the Forecast.Solar  API. Use read_series instead

        :raises NotImplementedError:
        """
        raise NotImplementedError("Close subscription is not implemented for Forecast.Solar.")

    async def _subscription_loop(
        self,
        handler: SubscriptionHandler,
        interval: TimeStep,
        req_interval: TimeStep,
        offset: TimeStep,
        data_interval: TimeStep,
    ) -> None:
        """
        .. warning::
            Cannot read single values from the Forecast.Solar  API. Use read_series instead

        :raises NotImplementedError:
        """
        raise NotImplementedError("Subscription loop is not implemented for Forecast.Solar.")

    def timestr_from_datetime(self, dt: datetime) -> str:
        """Create an Forecast.Solar compatible time string.

        :param dt: Datetime object to convert to string.
        :return: Forecast.Solar compatible time string.
        """

        return dt.isoformat(sep="T", timespec="seconds").replace(":", "%3A").replace("+", "%2B")

    def _raw_request(self, method: str, url: str, **kwargs: Any) -> requests.Response:
        """Perform Forecast.Solar request and handle possibly resulting errors.

        :param method: HTTP request method.
        :param endpoint: Endpoint for the request (server URI is added automatically).
        :param kwargs: Additional arguments for the request.
        """
        if self._api_key != "None":
            log.info("The api_key is None and only the public functions are available of the forecastsolar.api.")

        response = self._session.request(method, url, **kwargs)
        # Check for request errors
        response.raise_for_status()

        return response

    def _validate_nodes(self, nodes: Nodes | None) -> set[NodeForecastSolar]:  # type: ignore
        vnodes = super()._validate_nodes(nodes)
        _nodes = set()
        for node in vnodes:
            if isinstance(node, NodeForecastSolar):
                _nodes.add(node)

        return _nodes

    @classmethod
    def route_valid(cls, nodes: Nodes) -> bool:
        """Check if node routes make up a valid route, by using the Forecast.Solar API's check endpoint.

        :param nodes: List of nodes to check.
        :return: Boolean if the nodes are on the same route.
        """
        conn = ForecastSolarConnection()
        nodes = conn._validate_nodes(nodes)

        for node in nodes:
            _check_node = node.evolve(api_key=None, endpoint="check", data=None)
            try:
                conn._raw_request("GET", _check_node.url, headers=conn._headers)
            except requests.exceptions.HTTPError as e:
                log.error(f"\nRoute of node: {node.name} could not be verified: \n{e}")
                return False

        # If no request error occurred, the routes are valid
        return True

    @classmethod
    def calculate_watt_hours_period(cls, df: pd.DataFrame, watts_column: str) -> pd.DataFrame:
        """
        Calculates watt hours for each period based on the average watts provided.
        Assumes the DataFrame is indexed by timestamps.
        """
        # Calculate the duration of each period in hours
        df["period_hours"] = df.index.to_series().diff().dt.total_seconds() / 3600
        df["period_hours"].iloc[0] = df["period_hours"].mean()  # Handle the first NaN

        # Calculate watt_hours for each period
        df["watt_hours_period"] = df[watts_column] * df["period_hours"]
        return df.drop(columns=["period_hours"])

    @classmethod
    def summarize_watt_hours_over_day(cls, df: pd.DataFrame) -> pd.DataFrame:
        """
        Sums up watt hours over each day.
        """
        df["date"] = df.index.date
        daily_energy = df.groupby("date")["watt_hours_period"].sum().reset_index()
        daily_energy.columns = ["date", "watt_hours"]
        return daily_energy

    @classmethod
    def get_dataframe_of_values(
        cls, df: pd.DataFrame, watts_column: str = "watts"
    ) -> tuple[pd.DataFrame, pd.DataFrame]:
        """
        Process the original DataFrame to return a DataFrame with
        watt_hours_period, watt_hours (summarized over the day), and watt_hours_day.
        """
        watt_hours = cls.calculate_watt_hours_period(df, watts_column)
        daily_sum = cls.summarize_watt_hours_over_day(watt_hours)

        return watt_hours, daily_sum

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        tb: TracebackType | None,
    ) -> bool:
        if exc_type is not None:
            traceback.print_exception(exc_type, exc_value, tb)
            return False
        try:
            self._session.close()
        except Exception as e:
            log.error(f"Error closing the connection: {e}")
        return True

    def __del__(self) -> None:
        try:
            self._session.close()
        finally:
            pass



"""Utility functions for connecting to the ENTSO-E Transparency database and for reading data. This connector
does not have the ability to write data.
"""

from __future__ import annotations

import concurrent.futures
from datetime import datetime, timedelta, timezone
from typing import TYPE_CHECKING

import pandas as pd
import requests
from lxml import etree
from lxml.builder import E
from requests_cache import DO_NOT_CACHE, CachedSession

from eta_utility import get_logger
from eta_utility.connectors.node import NodeEntsoE
from eta_utility.timeseries import df_resample, df_time_slice
from eta_utility.util import dict_search, round_timestamp

if TYPE_CHECKING:
    from collections.abc import Mapping
    from typing import Any, Final

    from eta_utility.type_hints import Nodes, TimeStep

from .base_classes import SeriesConnection, SubscriptionHandler

log = get_logger("connectors.entso-e")


class ENTSOEConnection(SeriesConnection[NodeEntsoE], protocol="entsoe"):
    """
    ENTSOEConnection is a class to download and upload multiple features from and to the ENTSO-E transparency platform
    database as timeseries. The platform contains data about the european electricity markets.

    :param url: Url of the server with scheme (https://web-api.tp.entsoe.eu/)
    :param usr: Username for login to the platform (usually not required - default: None)
    :param pwd: Password for login to the platform (usually not required - default: None)
    :param api_token: Token for API authentication
    :param nodes: Nodes to select in connection
    """

    API_PATH: str = "/api"

    def __init__(
        self,
        url: str = "https://web-api.tp.entsoe.eu/",
        *,
        api_token: str,
        nodes: Nodes[NodeEntsoE] | None = None,
    ) -> None:
        url = url + self.API_PATH
        self._api_token: str = api_token
        super().__init__(url, None, None, nodes=nodes)

        self._node_ids: str | None = None
        self.config = _ConnectionConfiguration()
        self._session: CachedSession = CachedSession(
            cache_name="eta_utility/connectors/requests_cache/entso_e_cache",
            urls_expire_after={
                "https://web-api.tp.entsoe.eu/*": timedelta(minutes=15),
                "*": DO_NOT_CACHE,  # Don't cache other URLs
            },
            allowable_codes=(200, 400, 401),
            use_cache_dir=True,
        )

    @classmethod
    def _from_node(cls, node: NodeEntsoE, **kwargs: Any) -> ENTSOEConnection:
        """Initialize the connection object from an entso-e protocol node object

        :param node: Node to initialize from
        :param kwargs: Keyword arguments for API authentication, where "api_token" is required
        :return: ENTSOEConnection object
        """

        if "api_token" not in kwargs:
            raise AttributeError("Missing required function parameter api_token.")
        api_token = kwargs["api_token"]

        return super()._from_node(node, api_token=api_token)

    def read(self, nodes: Nodes[NodeEntsoE] | None = None) -> pd.DataFrame:
        """
        .. warning::
            Cannot read single values from ENTSO-E transparency platform. Use read_series instead

        :param nodes: List of nodes to read values from
        :return: Pandas DataFrame containing the data read from the connection
        """
        raise NotImplementedError(
            "Cannot read single values from ENTSO-E transparency platform. Use read_series instead"
        )

    def write(
        self, values: Mapping[NodeEntsoE, Mapping[datetime, Any]], time_interval: timedelta | None = None
    ) -> None:
        """
        .. warning::
            Cannot write to ENTSO-E transparency platform.

        :param values: Dictionary of nodes and data to write. {node: value}
        :param time_interval: Interval between datapoints (i.e. between "From" and "To" in EnEffCo Upload), default 1s
        """
        raise NotImplementedError("Cannot write to ENTSO-E transparency platform.")

    def subscribe(
        self, handler: SubscriptionHandler, nodes: Nodes[NodeEntsoE] | None = None, interval: TimeStep = 1
    ) -> None:
        """Subscribe to nodes and call handler when new data is available. This will return only the
        last available values.

        :param handler: SubscriptionHandler object with a push method that accepts node, value pairs
        :param interval: interval for receiving new data. It is interpreted as seconds when given as an integer.
        :param nodes: identifiers for the nodes to subscribe to
        """
        self.subscribe_series(handler=handler, req_interval=1, nodes=nodes, interval=interval, data_interval=interval)

    def _handle_xml(self, xml_content: bytes) -> dict[str, dict[str, list[pd.Series]]]:
        """Transform XML data from request response into dictionary containig resolutions and time series for the node.

        :param xml_content: XML data
        :return: Dictionary with resolutions and time series data
        """
        parser = etree.XMLParser(load_dtd=False, ns_clean=True, remove_pis=True)
        xml_data = etree.XML(xml_content, parser)
        ns = xml_data.nsmap
        data: dict[str, dict[str, list[pd.Series]]] = {}
        request_type = xml_data.find(".//type", namespaces=ns).text

        timeseries = xml_data.findall(".//TimeSeries", namespaces=ns)
        for ts in timeseries:
            # Day-Ahead Price
            if request_type == "A44":
                col_name = "Price"

            # Actual Generation per Type
            if request_type == "A75":
                psr_type = ts.find(".//MktPSRType", namespaces=ns).find("psrType", namespaces=ns).text
                col_name = dict_search(self.config.psr_types, psr_type)

                if ts.find(".//inBiddingZone_Domain.mRID", namespaces=ns) is not None:
                    col_name = col_name + "_Generation"
                elif ts.find(".//outBiddingZone_Domain.mRID", namespaces=ns) is not None:
                    col_name = col_name + "_Consumption"

            # contains the data points
            period = ts.find(".//Period", namespaces=ns)

            # datetime range of the data points
            time_interval = period.find(".//timeInterval", namespaces=ns).getchildren()
            resolution = period.find(".//resolution", namespaces=ns).text[2:4]  # truncating string PR60M
            datetime_range = pd.date_range(
                datetime.strptime(time_interval[0].text, "%Y-%m-%dT%H:%MZ"),
                datetime.strptime(time_interval[1].text, "%Y-%m-%dT%H:%MZ"),
                freq=resolution + "min",
                inclusive="left",
            )

            points = period.findall(".//Point", namespaces=ns)
            ts_data = [point.getchildren()[-1].text for point in points]

            s = pd.Series(data=ts_data, index=datetime_range, name=col_name)
            s.index = s.index.tz_localize(tz="UTC")  # ENTSO-E returns always UTC

            if resolution not in data:
                data[resolution] = {}

            if col_name not in data[resolution]:
                data[resolution][col_name] = []

            data[resolution][col_name].append(s.astype(float))

        return data

    def read_series(
        self,
        from_time: datetime,
        to_time: datetime,
        nodes: Nodes[NodeEntsoE] | None = None,
        interval: TimeStep = 1,
        **kwargs: Any,
    ) -> pd.DataFrame:
        """Download timeseries data from the ENTSO-E Database

        :param nodes: List of nodes to read values from
        :param from_time: Starting time to begin reading (included in output)
        :param to_time: Time to stop reading at (not included in output)
        :param interval: interval between time steps. It is interpreted as seconds if given as integer.
        :return: Pandas DataFrame containing the data read from the connection
        """
        nodes = self._validate_nodes(nodes)
        interval = interval if isinstance(interval, timedelta) else timedelta(seconds=interval)

        from_time = round_timestamp(from_time, interval.total_seconds())
        to_time = round_timestamp(to_time, interval.total_seconds())

        if from_time.tzinfo != to_time.tzinfo:
            log.warning(
                f"Timezone of from_time and to_time are different. Using from_time timezone: {from_time.tzinfo}"
            )

        def read_node(node: NodeEntsoE) -> pd.DataFrame:
            params = self.config.create_params(node, from_time, to_time)

            result = self._raw_request(params)
            data = self._handle_xml(result.content)

            df_dict = {}
            # All resolutions are resampled separatly and concatenated to one dataframe in the end
            for resolution in data:
                data_resolution = {
                    f"{node.name}_{column}": pd.concat(series) for column, series in data[resolution].items()
                }
                df_resolution = pd.DataFrame.from_dict(data_resolution, orient="columns")
                # entsoe always returns a dataframe in UTC time, convert to same time zone as given from_time
                df_resolution.index = df_resolution.index.tz_convert(tz=from_time.tzinfo)
                df_resolution = df_resample(df_resolution, interval, missing_data="fillna")
                df_resolution = df_time_slice(df_resolution, from_time, to_time)
                df_dict[resolution] = df_resolution

            value_df = pd.concat(df_dict.values(), axis=1, keys=df_dict.keys())
            return value_df.swaplevel(axis=1)

        with concurrent.futures.ThreadPoolExecutor() as executor:
            results = executor.map(read_node, nodes)

        return pd.concat(results, axis=1, sort=False)

    def subscribe_series(
        self,
        handler: SubscriptionHandler,
        req_interval: TimeStep,
        offset: TimeStep | None = None,
        nodes: Nodes[NodeEntsoE] | None = None,
        interval: TimeStep = 1,
        data_interval: TimeStep = 1,
        **kwargs: Any,
    ) -> None:
        """
        .. warning::
            Not implemented: Cannot subscribe to data from the ENTSO-E transparency platform.

        :param handler: SubscriptionHandler object with a push method that accepts node, value pairs
        :param req_interval: Duration covered by requested data (time interval). Interpreted as seconds if given as int
        :param offset: Offset from datetime.now from which to start requesting data (time interval).
                       Interpreted as seconds if given as int. Use negative values to go to past timestamps.
        :param data_interval: Time interval between values in returned data. Interpreted as seconds if given as int.
        :param interval: interval (between requests) for receiving new data.
                         It it interpreted as seconds when given as an integer.
        :param nodes: identifiers for the nodes to subscribe to
        """
        raise NotImplementedError("Cannot subscribe to data from the ENTSO-E transparency platform.")

    def close_sub(self) -> None:
        """
        .. warning::
            Not implemented: Cannot subscribe to data from the ENTSO-E transparency platform.
        """
        raise NotImplementedError("Cannot subscribe to data from the ENTSO-E transparency platform.")

    def _raw_request(self, params: Mapping[str, str], **kwargs: Mapping[str, Any]) -> requests.Response:
        """Perform ENTSO-E request and handle possibly resulting errors.

        :param params: Parameters to identify the endpoint
        :param kwargs: Additional arguments for the request.
        :return: request response
        """

        # Prepare the basic request for usage in the requests.
        headers = {"Content-Type": "application/xml", "SECURITY_TOKEN": self._api_token}

        xml = self.config.xml_head()
        for param, val in params.items():
            xml.append(self.config.xml_param(param, val))

        response = self._session.post(self.url, data=etree.tostring(xml), headers=headers, **kwargs)

        if response.status_code == 400:
            try:
                parser = etree.XMLParser(load_dtd=False, ns_clean=True, remove_pis=True)

                e_msg = etree.XML(response.content, parser)
                ns = e_msg.nsmap
                e_code = e_msg.find(".//Reason", namespaces=ns).find("code", namespaces=ns).text
                e_text = e_msg.find(".//Reason", namespaces=ns).find("text", namespaces=ns).text
                response.reason = f"ENTSO-E Error {response.status_code} ({e_code}: {e_text})"
            except Exception:
                pass

        response.raise_for_status()

        return response


class _ConnectionConfiguration:
    """Auxiliary class to configure the parameters for establishing a connection to ENTSO-E API.

    Currently, the connection class only supports two types of data requests through the method read_series, they are:
    **Energy price day ahead** and **Actual energy generation per type**. All the data requests available are listed in
    the _doc_type class attribute, but each of them contains a mandatory list of parameters to establish the connection,
    which can be seemed in the ENTSO-E documentation_.

    .. _documentation: https://transparency.entsoe.eu/content/static_content/Static%20content/web%20api/Guide.html
    """

    #: XML Namespace for the API
    _XMLNS: Final[str] = "urn:iec62325.351:tc57wg16:451-5:statusrequestdocument:4:0"
    #: bidding zones is a mapping of three letter iso country codes to bidding zones.
    _BIDDING_ZONES: Final[dict[str, str]] = {
        "DEU": "10Y1001A1001A83F",
        "DEU-AUT-LUX": "10Y1001A1001A63L",
        "ALB": "10YAL-KESH-----5",
        "AUT": "10YAT-APG------L",
        "BLR": "10Y1001A1001A51S",
        "BEL": "10YBE----------2",
        "BIH": "10YBA-JPCC-----D",
        "BGR": "10YCA-BULGARIA-R",
        "CZE-DEU-SVK": "10YDOM-CZ-DE-SKK",
        "HRV": "10YHR-HEP------M",
        "CYP": "10YCY-1001A0003J",
        "CZE": "10YCZ-CEPS-----N",
        "DEU-LUX": "10Y1001A1001A82H",
        "DNK_west": "10YDK-1--------W",
        "DNK_central": "10YDK-2--------M",
        "EST": "10Y1001A1001A39I",
        "FIN": "10YFI-1--------U",
        "MKD": "10YMK-MEPSO----8",
        "FRA": "10YFR-RTE------C",
        "GB": "17Y0000009369493",
        "GRC": "10YGR-HTSO-----Y",
        "HUN": "10YHU-MAVIR----U",
        "IRL": "10Y1001A1001A59C",
        "ITA_brindisi": "10Y1001A1001A699",
        "ITA_calabria": "10Y1001C--00096J",
        "ITA_central_north": "10Y1001A1001A70O",
        "ITA_central_south": "10Y1001A1001A71M",
        "ITA_foggia": "10Y1001A1001A72K",
        "ITA-GRC": "10Y1001A1001A66F",
        "ITA_malta": "10Y1001A1001A877",
        "ITA_north": "10Y1001A1001A73I",
        "ITA-AUT": "10Y1001A1001A80L",
        "ITA-CHE": "10Y1001A1001A68B",
        "ITA-FRA": "10Y1001A1001A81J",
        "ITA-SVN": "10Y1001A1001A67D",
        "ITA_priolo": "10Y1001A1001A76C",
        "ITA_rossano": "10Y1001A1001A77A",
        "ITA_sardinia": "10Y1001A1001A74G",
        "ITA_sicily": "10Y1001A1001A75E",
        "ITA_south": "10Y1001A1001A788",
        "RUS_kaliningrad": "10Y1001A1001A50U",
        "LVA": "10YLV-1001A00074",
        "LTU": "10YLT-1001A0008Q",
        "LUX": "10YLU-CEGEDEL-NQ",
        "MLT": "10Y1001A1001A93C",
        "MNE": "10YCS-CG-TSO---S",
        "GBR": "10YGB----------A",
        "NLD": "10YNL----------L",
        "NOR_1": "10YNO-1--------2",
        "NOR_2": "10YNO-2--------T",
        "NOR_3": "10YNO-3--------J",
        "NOR_4": "10YNO-4--------9",
        "NOR_5": "10Y1001A1001A48H",
        "POL": "10YPL-AREA-----S",
        "PRT": "10YPT-REN------W",
        "MDA": "10Y1001A1001A990",
        "ROU": "10YRO-TEL------P",
        "RUS": "10Y1001A1001A49F",
        "SWE_1": "10Y1001A1001A44P",
        "SWE_2": "10Y1001A1001A45N",
        "SWE_3": "10Y1001A1001A46L",
        "SWE_4": "10Y1001A1001A47J",
        "SRB": "10YCS-SERBIATSOV",
        "SVK": "10YSK-SEPS-----K",
        "SVN": "10YSI-ELES-----O",
        "ESP": "10YES-REE------0",
        "SWE": "10YSE-1--------K",
        "CHE": "10YCH-SWISSGRIDZ",
        "TUR": "10YTR-TEIAS----W",
        "UKR": "10Y1001C--00003F",
    }

    _MARKET_AGREEMENTS: Final[dict[str, str]] = {
        "Daily": "A01",
        "Weekly": "A02",
        "Monthly": "A03",
        "Yearly": "A04",
        "Total": "A05",
        "Long term": "A06",
        "Intraday": "A07",
        "Hourly": "A13",
    }

    _AUCTION_TYPES: Final[dict[str, str]] = {
        "Implicit": "A01",
        "Explicit": "A02",
    }

    _AUCTION_CATEGORIES: Final[dict[str, str]] = {
        "Base": "A01",
        "Peak": "A02",
        "Off Peak": "A03",
        "Hourly": "A04",
    }

    _PSR_TYPES: Final[dict[str, str]] = {
        "Mixed": "A03",
        "Generation": "A04",
        "Load": "A05",
        "Biomass": "B01",
        "Fossil Brown coal/Lignite": "B02",
        "Fossil Coal-derived gas": "B03",
        "Fossil Gas": "B04",
        "Fossil Hard coal": "B05",
        "Fossil Oil": "B06",
        "Fossil Oil shale": "B07",
        "Fossil Peat": "B08",
        "Geothermal": "B09",
        "Hydro Pumped Storage": "B10",
        "Hydro Run-of-river and poundage": "B11",
        "Hydro Water Reservoir": "B12",
        "Marine": "B13",
        "Nuclear": "B14",
        "Other renewable": "B15",
        "Solar": "B16",
        "Waste": "B17",
        "Wind Offshore": "B18",
        "Wind Onshore": "B19",
        "Other": "B20",
        "AC Link": "B21",
        "DC Link": "B22",
        "Substation": "B23",
        "Transformer": "B24",
    }

    _BUSINESS_TYPES: Final[dict[str, str]] = {
        "General Capacity Information": "A25",
        "Already allocated capacity (AAC)": "A29",
        "Requested capacity (without price)": "A43",
        "System Operator redispatching": "A46",
        "Planned maintenance": "A53",
        "Unplanned outage": "A54",
        "Internal redispatch": "A85",
        "Frequency containment reserve": "A95",
        "Automatic frequency restoration reserve": "A96",
        "Manual frequency restoration reserve": "A97",
        "Replacement reserve": "A98",
        "Interconnector network evolution": "B01",
        "Interconnector network dismantling": "B02",
        "Counter trade": "B03",
        "Congestion costs": "B04",
        "Capacity allocated (including price)": "B05",
        "Auction revenue": "B07",
        "Total nominated capacity": "B08",
        "Net position": "B09",
        "Congestion income": "B10",
        "Production unit": "B11",
        "Area Control Error": "B33",
        "Procured capacity": "B95",
        "Shared Balancing Reserve Capacity": "C22",
        "Share of reserve capacity": "C23",
        "Actual reserve capacity": "C24",
    }

    _PROCESS_TYPES: Final[dict[str, str]] = {
        "Day ahead": "A01",
        "Intra day incremental": "A02",
        "Realised": "A16",
        "Intraday total": "A18",
        "Week ahead": "A31",
        "Month ahead": "A32",
        "Year ahead": "A33",
        "Synchronisation process": "A39",
        "Intraday process": "A40",
        "Replacement reserve": "A46",
        "Manual frequency restoration reserve": "A47",
        "Automatic frequency restoration reserve": "A51",
        "Frequency containment reserve": "A52",
        "Frequency restoration reserve": "A56",
    }

    _DOC_STATES: Final[dict[str, str]] = {
        "Intermediate": "A01",
        "Final": "A02",
        "Active": "A05",
        "Cancelled": "A09",
        "Withdrawn": "A13",
        "Estimated": "X01",
    }

    _DOC_TYPES: Final[dict[str, str]] = {
        "FinalisedSchedule": "A09",
        "AggregatedEnergyDataReport": "A11",
        "AcquiringSystemOperatorReserveSchedule": "A15",
        "Bid": "A24",
        "AllocationResult": "A25",
        "Capacity": "A26",
        "AgreedCapacity": "A31",
        "ReserveAllocationResult": "A38",
        "Price": "A44",
        "EstimatedNetTransferCapacity": "A61",
        "RedispatchNotice": "A63",
        "SystemTotalLoad": "A65",
        "InstalledGenerationPerType": "A68",
        "WindAndSolarForecast": "A69",
        "LoadForecastMargin": "A70",
        "GenerationForecast": "A71",
        "ReservoirFillingInformation": "A72",
        "ActualGeneration": "A73",
        "WindAndSolarGeneration": "A74",
        "ActualGenerationPerType": "A75",
        "LoadUnavailability": "A76",
        "ProductionUnavailability": "A77",
        "TransmissionUnavailability": "A78",
        "OffshoreGridInfrastructureUnavailability": "A79",
        "GenerationUnavailability": "A80",
        "ContractedReserves": "A81",
        "AcceptedOffers": "A82",
        "ActivatedBalancingQuantities": "A83",
        "ActivatedBalancingPrices": "A84",
        "ImbalancePrices": "A85",
        "ImbalanceVolume": "A86",
        "FinancialSituation": "A87",
        "CrossBorderBalancing": "A88",
        "ContractedReservePrices": "A89",
        "InterconnectionNetworkExpansion": "A90",
        "CounterTradeNotice": "A91",
        "CongestionCosts": "A92",
        "DCLinkCapacity": "A93",
        "NonEUAllocations": "A94",
        "Configuration": "A95",
        "FlowBasedAllocations": "B11",
    }

    def create_params(self, node: NodeEntsoE, from_time: datetime, to_time: datetime) -> dict[str, str]:
        """Create request parameters object according to API specifications
        Handle configuration paramters for each type of connection

        :param node: ENTSO-E Node
        :param from_time: Starting time
        :param to_time: End time
        :return: Dictionary with parameters
        """
        if node.endpoint not in self._DOC_TYPES:
            raise ValueError(f"Unsupported endpoint for ENTSO-E connection: {node.endpoint}.")

        params = {"DocumentType": node.endpoint}
        if node.endpoint == "ActualGenerationPerType":
            params["ProcessType"] = "Realised"
            params["In_Domain"] = node.bidding_zone

        elif node.endpoint == "Price":
            params["ProcessType"] = "Day ahead"
            params["In_Domain"] = node.bidding_zone
            params["Out_Domain"] = node.bidding_zone

        else:
            raise NotImplementedError(f"Endpoint not available: {node.endpoint}")

        # Round down at from_time and up at to_time to receive all necessary values from entsoe
        # entsoe uses always a full hour
        rounded_from_time_utc = round_timestamp(from_time.astimezone(timezone.utc), 3600) - timedelta(hours=1)
        rounded_to_time_utc = round_timestamp(to_time.astimezone(timezone.utc), 3600)

        params["TimeInterval"] = (
            f"{rounded_from_time_utc.strftime('%Y-%m-%dT%H:%M:%SZ')}/"
            f"{rounded_to_time_utc.strftime('%Y-%m-%dT%H:%M:%SZ')}"
        )
        return params

    def xml_head(self) -> etree.ElementTree:
        """Create header of the xml data for the POST method.

        :return: tree of elements with the pre-defined values for the request
        """
        now = datetime.utcnow()
        # Prepare XML Header data
        data = E("StatusRequest_MarketDocument", xmlns=self._XMLNS)
        data.append(E("mRID", f"Request_{now.isoformat(sep='T', timespec='seconds')}"))
        data.append(E("type", "A59"))
        data.append(E("sender_MarketParticipant.mRID", "10X1001A1001A450", codingScheme="A01"))
        data.append(E("sender_MarketParticipant.marketRole.type", "A07"))
        data.append(E("receiver_MarketParticipant.mRID", "10X1001A1001A450", codingScheme="A01"))
        data.append(E("receiver_MarketParticipant.marketRole.type", "A32"))
        data.append(E("createdDateTime", f"{now.isoformat(sep='T', timespec='seconds')}Z"))

        return data

    def xml_param(self, parameter: str, value: str) -> etree.Element:
        """Map parameters to request values for the xml document.

        :return: tree with parameters
        """
        if parameter in {"Contract_MarketAgreement.Type", "Type_MarketAgreement.Type"}:
            value = self._MARKET_AGREEMENTS[value]
        elif parameter == "Auction.Type":
            value = self._AUCTION_TYPES[value]
        elif parameter == "Auction.Category":
            value = self._AUCTION_CATEGORIES[value]
        elif parameter == "PsrType":
            value = self._PSR_TYPES[value]
        elif parameter == "BusinessType":
            value = self._BUSINESS_TYPES[value]
        elif parameter == "ProcessType":
            value = self._PROCESS_TYPES[value]
        elif parameter == "DocStatus":
            value = self._DOC_STATES[value]
        elif parameter == "DocumentType":
            value = self._DOC_TYPES[value]
        elif parameter in {"In_Domain", "Out_Domain"}:
            value = self._BIDDING_ZONES[value]

        return E("AttributeInstanceComponent", E("attribute", parameter), E("attributeValue", value))

    @property
    def psr_types(self) -> dict[str, str]:
        return self._PSR_TYPES

    @property
    def doc_types(self) -> dict[str, str]:
        return self._DOC_TYPES
